# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""
Registration workflows
++++++++++++++++++++++

.. autofunction:: init_bold_reg_wf
.. autofunction:: init_bold_t1_trans_wf
.. autofunction:: init_bbreg_wf
.. autofunction:: init_fsl_bbr_wf

"""

import logging
import os
import typing as ty

from nipype.interfaces import c3, fsl
from nipype.interfaces import utility as niu
from nipype.pipeline import engine as pe

from nibabies import config, data
from nibabies._types import AffineDOF, RegistrationInit

DEFAULT_MEMORY_MIN_GB = config.DEFAULT_MEMORY_MIN_GB
LOGGER = logging.getLogger('nipype.workflow')


def init_bold_reg_wf(
    freesurfer: bool,
    use_bbr: bool,
    bold2anat_dof: AffineDOF,
    bold2anat_init: RegistrationInit,
    mem_gb: float,
    omp_nthreads: int,
    name: str = 'bold_reg_wf',
    sloppy: bool = False,
):
    """
    Build a workflow to run same-subject, BOLD-to-anat image-registration.

    Calculates the registration between a reference BOLD image and anatomical-space
    using a boundary-based registration (BBR) cost function.
    If FreeSurfer-based preprocessing is enabled, the ``bbregister`` utility
    is used to align the BOLD images to the reconstructed subject, and the
    resulting transform is adjusted to target the anatomical space.
    If FreeSurfer-based preprocessing is disabled, FSL FLIRT is used with the
    BBR cost function to directly target the T1 space.

    Workflow Graph
        .. workflow::
            :graph2use: orig
            :simple_form: yes

            from fmriprep.workflows.bold.registration import init_bold_reg_wf
            wf = init_bold_reg_wf(freesurfer=True,
                                  mem_gb=3,
                                  omp_nthreads=1,
                                  use_bbr=True,
                                  bold2anat_dof=9,
                                  bold2anat_init='auto')

    Parameters
    ----------
    freesurfer : :obj:`bool`
        Enable FreeSurfer functional registration (bbregister)
    use_bbr : :obj:`bool` or None
        Enable/disable boundary-based registration refinement.
        If ``None``, test BBR result for distortion before accepting.
    bold2anat_dof : 6, 9 or 12
        Degrees-of-freedom for BOLD-anatomical registration
    bold2anat_init : str, 't1w', 't2w' or 'header'
        If ``'header'``, use header information for initialization of BOLD and T1 images.
        If ``'t1w'``, align BOLD to T1w by their centers.
        If ``'t2w'``, align BOLD to T1w using the T2w as an intermediate.
    mem_gb : :obj:`float`
        Size of BOLD file in GB
    omp_nthreads : :obj:`int`
        Maximum number of threads an individual process may use
    name : :obj:`str`
        Name of workflow (default: ``bold_reg_wf``)

    Inputs
    ------
    ref_bold_brain
        Reference image to which BOLD series is aligned
        If ``fieldwarp == True``, ``ref_bold_brain`` should be unwarped
    anat_preproc
        Anatomical reference volume
    anat_brain
        Skull-stripped ``anat_preproc``
    anat_dseg
        Segmentation of preprocessed structural image, including
        gray-matter (GM), white-matter (WM) and cerebrospinal fluid (CSF)
    subjects_dir
        FreeSurfer SUBJECTS_DIR
    subject_id
        FreeSurfer subject ID
    fsnative2anat_xfm
        LTA-style affine matrix translating from FreeSurfer-conformed subject space to anatomical

    Outputs
    -------
    itk_bold_to_anat
        Affine transform from ``ref_bold_brain`` to anatomical space (ITK format)
    itk_anat_to_bold
        Affine transform from anatomical space to BOLD space (ITK format)
    fallback
        Boolean indicating whether BBR was rejected (mri_coreg registration returned)

    See Also
    --------
      * :py:func:`~fmriprep.workflows.bold.registration.init_bbreg_wf`
      * :py:func:`~fmriprep.workflows.bold.registration.init_fsl_bbr_wf`

    """
    from niworkflows.engine.workflows import LiterateWorkflow as Workflow

    workflow = Workflow(name=name)
    inputnode = pe.Node(
        niu.IdentityInterface(
            fields=[
                'ref_bold_brain',
                'anat_preproc',
                'anat_mask',
                'anat_dseg',
                'subjects_dir',
                'subject_id',
                'fsnative2anat_xfm',
            ]
        ),
        name='inputnode',
    )

    outputnode = pe.Node(
        niu.IdentityInterface(fields=['itk_bold_to_anat', 'itk_anat_to_bold', 'fallback']),
        name='outputnode',
    )

    if freesurfer:
        bbr_wf = init_bbreg_wf(
            use_bbr=use_bbr,
            bold2anat_dof=bold2anat_dof,
            bold2anat_init=bold2anat_init,
            omp_nthreads=omp_nthreads,
        )
    else:
        bbr_wf = init_fsl_bbr_wf(
            use_bbr=use_bbr,
            bold2anat_dof=bold2anat_dof,
            bold2anat_init=bold2anat_init,
            sloppy=sloppy,
            omp_nthreads=omp_nthreads,
        )

    # fmt: off
    workflow.connect([
        (inputnode, bbr_wf, [
            ('ref_bold_brain', 'inputnode.in_file'),
            ('fsnative2anat_xfm', 'inputnode.fsnative2anat_xfm'),
            ('subjects_dir', 'inputnode.subjects_dir'),
            ('subject_id', 'inputnode.subject_id'),
            ('anat_preproc', 'inputnode.anat_preproc'),
            ('anat_mask', 'inputnode.anat_mask'),
            ('anat_dseg', 'inputnode.anat_dseg'),
        ]),
        (bbr_wf, outputnode, [
            ('outputnode.itk_bold_to_anat', 'itk_bold_to_anat'),
            ('outputnode.itk_anat_to_bold', 'itk_anat_to_bold'),
            ('outputnode.fallback', 'fallback'),
        ]),
    ])  # fmt:skip

    return workflow


def init_bbreg_wf(
    use_bbr: bool,
    bold2anat_dof: AffineDOF,
    bold2anat_init: RegistrationInit,
    omp_nthreads: int,
    name: str = 'bbreg_wf',
):
    """
    Build a workflow to run FreeSurfer's ``bbregister``.

    This workflow uses FreeSurfer's ``bbregister`` to register a BOLD image to
    a T2-weighted or T1-weighted structural image.

    It is a counterpart to :py:func:`~fmriprep.workflows.bold.registration.init_fsl_bbr_wf`,
    which performs the same task using FSL's FLIRT with a BBR cost function.
    The ``use_bbr`` option permits a high degree of control over registration.
    If ``False``, standard, affine coregistration will be performed using
    FreeSurfer's ``mri_coreg`` tool.
    If ``True``, ``bbregister`` will be seeded with the initial transform found
    by ``mri_coreg`` (equivalent to running ``bbregister --init-coreg``).
    If ``None``, after ``bbregister`` is run, the resulting affine transform
    will be compared to the initial transform found by ``mri_coreg``.
    Excessive deviation will result in rejecting the BBR refinement and
    accepting the original, affine registration.

    Workflow Graph
        .. workflow ::
            :graph2use: orig
            :simple_form: yes

            from nibabies.workflows.bold.registration import init_bbreg_wf
            wf = init_bbreg_wf(use_bbr=True, bold2anat_dof=9,
                               bold2anat_init='t1w', omp_nthreads=1)


    Parameters
    ----------
    use_bbr : :obj:`bool` or None
        Enable/disable boundary-based registration refinement.
        If ``None``, test BBR result for distortion before accepting.
    bold2anat_dof : 6, 9 or 12
        Degrees-of-freedom for BOLD-anatomical registration
    bold2anat_init : str, 't1w', 't2w' or 'header'
        If ``'header'``, use header information for initialization of BOLD and T1 images.
        If ``'t1w'``, align BOLD to T1w by their centers.
        If ``'t2w'``, align BOLD to T1w using the T2w as an intermediate.
    name : :obj:`str`, optional
        Workflow name (default: bbreg_wf)

    Inputs
    ------
    in_file
        Reference BOLD image to be registered
    fsnative2anat_xfm
        FSL-style affine matrix translating from either
        FreeSurfer T1.mgz to T1w or T2.mgz to T2w
    subjects_dir
        FreeSurfer SUBJECTS_DIR
    subject_id
        FreeSurfer subject ID (must have folder in SUBJECTS_DIR)
    anat_preproc
        Unused (see :py:func:`~fmriprep.workflows.bold.registration.init_fsl_bbr_wf`)
    anat_mask
        Unused (see :py:func:`~fmriprep.workflows.bold.registration.init_fsl_bbr_wf`)
    anat_dseg
        Unused (see :py:func:`~fmriprep.workflows.bold.registration.init_fsl_bbr_wf`)

    Outputs
    -------
    itk_bold_to_anat
        Affine transform from ``ref_bold_brain`` to anatomical space (ITK format)
    itk_anat_to_bold
        Affine transform from anatomical space to BOLD space (ITK format)
    fallback
        Boolean indicating whether BBR was rejected (mri_coreg registration returned)

    """
    from nipype.interfaces.freesurfer import BBRegister
    from niworkflows.engine.workflows import LiterateWorkflow as Workflow
    from niworkflows.interfaces.nitransforms import ConcatenateXFMs
    from niworkflows.interfaces.patches import FreeSurferSource

    from nibabies.interfaces.patches import MRICoreg

    workflow = Workflow(name=name)
    workflow.__desc__ = """\
The BOLD reference was then co-registered to the anatomical reference using
`bbregister` (FreeSurfer) which implements boundary-based registration [@bbr].
Co-registration was configured with {dof} degrees of freedom{reason}.
""".format(
        dof={6: 'six', 9: 'nine', 12: 'twelve'}[bold2anat_dof],
        reason=(
            ''
            if bold2anat_dof == 6
            else 'to account for distortions remaining in the BOLD reference'
        ),
    )

    use_t2w = bold2anat_init == 't2w'
    if use_t2w:
        workflow.__desc__ += ' The aligned T2w image was used for initial co-registration.'

    inputnode = pe.Node(
        niu.IdentityInterface(
            [
                'in_file',
                'fsnative2anat_xfm',  # BBRegister
                'subjects_dir',
                'subject_id',
                'anat_preproc',  # FLIRT BBR
                'anat_mask',
                'anat_dseg',
            ]
        ),
        name='inputnode',
    )
    outputnode = pe.Node(
        niu.IdentityInterface(['itk_bold_to_anat', 'itk_anat_to_bold', 'fallback']),
        name='outputnode',
    )

    if bold2anat_init not in ty.get_args(RegistrationInit):
        raise ValueError(f'Unknown BOLD-to-anatomical initialization option: {bold2anat_init}')

    # For now make BBR unconditional - in the future, we can fall back to identity,
    # but adding the flexibility without testing seems a bit dangerous
    if bold2anat_init == 'header':
        if use_bbr is False:
            raise ValueError('Cannot disable BBR and use header registration')
        if use_bbr is None:
            LOGGER.warning('Initializing BBR with header; affine fallback disabled')
            use_bbr = True

    fssource = pe.Node(FreeSurferSource(), name='fssource')

    mri_coreg = pe.Node(
        MRICoreg(dof=bold2anat_dof, sep=[4], ftol=0.0001, linmintol=0.01),
        name='mri_coreg',
        n_procs=omp_nthreads,
        mem_gb=5,
    )
    if use_t2w:
        mri_coreg.inputs.reference_mask = False

    bbregister = pe.Node(
        BBRegister(
            dof=bold2anat_dof,
            contrast_type='t2',
            out_lta_file=True,
        ),
        name='bbregister',
        mem_gb=12,
    )
    if bold2anat_init == 'header':
        bbregister.inputs.init = 'header'

    transforms = pe.Node(niu.Merge(2), run_without_submitting=True, name='transforms')
    # In cases where Merge(2) only has `in1` or `in2` defined
    # output list will just contain a single element
    select_transform = pe.Node(
        niu.Select(index=0), run_without_submitting=True, name='select_transform'
    )
    merge_ltas = pe.Node(niu.Merge(2), name='merge_ltas', run_without_submitting=True)
    concat_xfm = pe.Node(ConcatenateXFMs(inverse=True), name='concat_xfm')

    workflow.connect([
        (inputnode, merge_ltas, [('fsnative2anat_xfm', 'in2')]),
        # Wire up the co-registration alternatives
        (transforms, select_transform, [('out', 'inlist')]),
        (select_transform, merge_ltas, [('out', 'in1')]),
        (merge_ltas, concat_xfm, [('out', 'in_xfms')]),
        (concat_xfm, outputnode, [('out_xfm', 'itk_bold_to_anat')]),
        (concat_xfm, outputnode, [('out_inv', 'itk_anat_to_bold')]),
    ])  # fmt:skip

    # Do not initialize with header, use mri_coreg
    if bold2anat_init != 'header':
        workflow.connect([
            (inputnode, mri_coreg, [('subjects_dir', 'subjects_dir'),
                                    ('subject_id', 'subject_id'),
                                    ('in_file', 'source_file')]),
            (mri_coreg, transforms, [('out_lta_file', 'in2')]),
        ])  # fmt:skip

        if use_t2w:
            workflow.connect([
                (inputnode, fssource, [('subjects_dir', 'subjects_dir'),
                                       ('subject_id', 'subject_id')]),
                (fssource, mri_coreg, [('T2', 'reference_file')]),
            ])  # fmt:skip

        # Short-circuit workflow building, use initial registration
        if use_bbr is False:
            outputnode.inputs.fallback = True

            return workflow

        # Otherwise bbregister will also be used
        workflow.connect(mri_coreg, 'out_lta_file', bbregister, 'init_reg_file')

    # Use bbregister
    workflow.connect([
        (inputnode, bbregister, [('subjects_dir', 'subjects_dir'),
                                 ('subject_id', 'subject_id'),
                                 ('in_file', 'source_file')]),
        (bbregister, transforms, [('out_lta_file', 'in1')]),
    ])  # fmt:skip

    # Short-circuit workflow building, use boundary-based registration
    if use_bbr is True:
        outputnode.inputs.fallback = False

        return workflow

    # Only reach this point if bold2anat_init is "t1w" or "t2w" and use_bbr is None
    compare_transforms = pe.Node(niu.Function(function=compare_xforms), name='compare_transforms')

    workflow.connect([
        (transforms, compare_transforms, [('out', 'lta_list')]),
        (compare_transforms, outputnode, [('out', 'fallback')]),
        (compare_transforms, select_transform, [('out', 'index')]),
    ])  # fmt:skip

    return workflow


def init_fsl_bbr_wf(
    use_bbr: bool,
    bold2anat_dof: AffineDOF,
    bold2anat_init: RegistrationInit,
    omp_nthreads: int,
    sloppy: bool = False,
    name: str = 'fsl_bbr_wf',
):
    """
    Build a workflow to run FSL's ``flirt``.

    This workflow uses FSL FLIRT to register a BOLD image to an anatomical
    structural image, using a boundary-based registration (BBR) cost function.
    It is a counterpart to :py:func:`~fmriprep.workflows.bold.registration.init_bbreg_wf`,
    which performs the same task using FreeSurfer's ``bbregister``.

    The ``use_bbr`` option permits a high degree of control over registration.
    If ``False``, standard, rigid coregistration will be performed by FLIRT.
    If ``True``, FLIRT-BBR will be seeded with the initial transform found by
    the rigid coregistration.
    If ``None``, after FLIRT-BBR is run, the resulting affine transform
    will be compared to the initial transform found by FLIRT.
    Excessive deviation will result in rejecting the BBR refinement and
    accepting the original, affine registration.

    Workflow Graph
        .. workflow ::
            :graph2use: orig
            :simple_form: yes

            from fmriprep.workflows.bold.registration import init_fsl_bbr_wf
            wf = init_fsl_bbr_wf(
                use_bbr=True, bold2anat_dof=9, bold2anat_init='t1w', omp_nthreads=1
            )


    Parameters
    ----------
    use_bbr : :obj:`bool` or None
        Enable/disable boundary-based registration refinement.
        If ``None``, test BBR result for distortion before accepting.
    bold2anat_dof : 6, 9 or 12
        Degrees-of-freedom for BOLD-anatomical registration
    bold2anat_init : str, 't1w', 't2w' or 'header'
        If ``'header'``, use header information for initialization of BOLD and T1 images.
        If ``'t1w'``, align BOLD to T1w by their centers.
        If ``'t2w'``, align BOLD to T1w using the T2w as an intermediate.
    name : :obj:`str`, optional
        Workflow name (default: fsl_bbr_wf)

    Inputs
    ------
    in_file
        Reference BOLD image to be registered
    anat_preproc
        Anatomical structural image
    anat_mask
        Brain mask of structural image
    anat_dseg
        FAST segmentation of masked ``t1w_preproc``
    fsnative2anat_xfm
        Unused (see :py:func:`~fmriprep.workflows.bold.registration.init_bbreg_wf`)
    subjects_dir
        Unused (see :py:func:`~fmriprep.workflows.bold.registration.init_bbreg_wf`)
    subject_id
        Unused (see :py:func:`~fmriprep.workflows.bold.registration.init_bbreg_wf`)

    Outputs
    -------
    itk_bold_to_anat
        Affine transform from ``ref_bold_brain`` to anatomical space (ITK format)
    itk_anat_to_bold
        Affine transform from anatomical space to BOLD space (ITK format)
    fallback
        Boolean indicating whether BBR was rejected (rigid FLIRT registration returned)

    """
    from nipype.interfaces.freesurfer import MRICoreg
    from niworkflows.engine.workflows import LiterateWorkflow as Workflow
    from niworkflows.interfaces.freesurfer import PatchedLTAConvert as LTAConvert
    from niworkflows.interfaces.nibabel import ApplyMask
    from niworkflows.utils.images import dseg_label as _dseg_label

    workflow = Workflow(name=name)
    workflow.__desc__ = """\
The BOLD reference was then co-registered to the anatomical reference using
`mri_coreg` (FreeSurfer) followed by `flirt` [FSL {fsl_ver}, @flirt]
with the boundary-based registration [@bbr] cost-function.
Co-registration was configured with {dof} degrees of freedom{reason}.
""".format(
        fsl_ver=fsl.FLIRT().version or '<ver>',
        dof={6: 'six', 9: 'nine', 12: 'twelve'}[bold2anat_dof],
        reason=(
            ''
            if bold2anat_dof == 6
            else 'to account for distortions remaining in the BOLD reference'
        ),
    )

    inputnode = pe.Node(
        niu.IdentityInterface(
            [
                'in_file',
                'fsnative2anat_xfm',  # BBRegister
                'subjects_dir',
                'subject_id',
                'anat_preproc',  # FLIRT BBR
                'anat_mask',
                'anat_dseg',
            ]
        ),
        name='inputnode',
    )
    outputnode = pe.Node(
        niu.IdentityInterface(['itk_bold_to_anat', 'itk_anat_to_bold', 'fallback']),
        name='outputnode',
    )

    # TODO: This may change on whether T1/T2 is used as reference
    wm_mask = pe.Node(niu.Function(function=_dseg_label), name='wm_mask')
    wm_mask.inputs.label = 2  # BIDS default is WM=2

    if bold2anat_init not in ty.get_args(RegistrationInit):
        raise ValueError(f'Unknown BOLD-anat initialization option: {bold2anat_init}')

    if bold2anat_init == 'header':
        raise NotImplementedError('Header-based registration initialization not supported for FSL')

    if bold2anat_init == 't2w':
        LOGGER.warning(
            'T2w intermediate for FSL is not implemented, registering with T1w instead.'
        )

    # Mask T1w_preproc with T1w_mask to make T1w_brain
    mask_anat_brain = pe.Node(ApplyMask(), name='mask_anat_brain')

    mri_coreg = pe.Node(
        MRICoreg(dof=bold2anat_dof, sep=[4], ftol=0.0001, linmintol=0.01),
        name='mri_coreg',
        n_procs=omp_nthreads,
        mem_gb=5,
    )

    lta_to_fsl = pe.Node(LTAConvert(out_fsl=True), name='lta_to_fsl', mem_gb=DEFAULT_MEMORY_MIN_GB)

    invt_bbr = pe.Node(
        fsl.ConvertXFM(invert_xfm=True), name='invt_bbr', mem_gb=DEFAULT_MEMORY_MIN_GB
    )

    # BOLD to anat transform matrix is from fsl, using c3 tools to convert to
    # something ANTs will like.
    fsl2itk_fwd = pe.Node(
        c3.C3dAffineTool(fsl2ras=True, itk_transform=True),
        name='fsl2itk_fwd',
        mem_gb=DEFAULT_MEMORY_MIN_GB,
    )
    fsl2itk_inv = pe.Node(
        c3.C3dAffineTool(fsl2ras=True, itk_transform=True),
        name='fsl2itk_inv',
        mem_gb=DEFAULT_MEMORY_MIN_GB,
    )
    # fmt:off
    workflow.connect([
        (inputnode, mask_anat_brain, [('anat_preproc', 'in_file'),
                                      ('anat_mask', 'in_mask')]),
        (inputnode, mri_coreg, [('in_file', 'source_file')]),
        (inputnode, fsl2itk_fwd, [('in_file', 'source_file')]),
        (inputnode, fsl2itk_inv, [('in_file', 'reference_file')]),
        (mask_anat_brain, mri_coreg, [('out_file', 'reference_file')]),
        (mask_anat_brain, fsl2itk_fwd, [('out_file', 'reference_file')]),
        (mask_anat_brain, fsl2itk_inv, [('out_file', 'source_file')]),
        (mri_coreg, lta_to_fsl, [('out_lta_file', 'in_lta')]),
        (invt_bbr, fsl2itk_inv, [('out_file', 'transform_file')]),
        (fsl2itk_fwd, outputnode, [('itk_transform', 'itk_bold_to_anat')]),
        (fsl2itk_inv, outputnode, [('itk_transform', 'itk_anat_to_bold')]),
    ])  # fmt:skip

    # Short-circuit workflow building, use rigid registration
    if use_bbr is False:
        # fmt:off
        workflow.connect([
            (lta_to_fsl, invt_bbr, [('out_fsl', 'in_file')]),
            (lta_to_fsl, fsl2itk_fwd, [('out_fsl', 'transform_file')]),
        ])
        # fmt:on
        outputnode.inputs.fallback = True

        return workflow

    flt_bbr = pe.Node(
        fsl.FLIRT(cost_func='bbr', dof=bold2anat_dof, args='-basescale 1'),
        name='flt_bbr',
    )

    FSLDIR = os.getenv('FSLDIR')
    if FSLDIR and os.path.exists(schedule := os.path.join(FSLDIR, 'etc/flirtsch/bbr.sch')):
        flt_bbr.inputs.schedule = schedule
    else:
        # Should mostly be hit while building docs
        LOGGER.warning('FSLDIR unset - using packaged BBR schedule')
        flt_bbr.inputs.schedule = data.load('flirtsch/bbr.sch')
    # fmt:off
    workflow.connect([
        (inputnode, wm_mask, [('anat_dseg', 'in_seg')]),
        (inputnode, flt_bbr, [('in_file', 'in_file')]),
        (lta_to_fsl, flt_bbr, [('out_fsl', 'in_matrix_file')]),
    ])
    # fmt:on
    if sloppy is True:
        downsample = pe.Node(
            niu.Function(
                function=_conditional_downsampling, output_names=['out_file', 'out_mask']
            ),
            name='downsample',
        )
        # fmt:off
        workflow.connect([
            (mask_anat_brain, downsample, [('out_file', 'in_file')]),
            (wm_mask, downsample, [('out', 'in_mask')]),
            (downsample, flt_bbr, [('out_file', 'reference'),
                                   ('out_mask', 'wm_seg')]),
        ])
        # fmt:on
    else:
        # fmt:off
        workflow.connect([
            (mask_anat_brain, flt_bbr, [('out_file', 'reference')]),
            (wm_mask, flt_bbr, [('out', 'wm_seg')]),
        ])
        # fmt:on

    # Short-circuit workflow building, use boundary-based registration
    if use_bbr is True:
        # fmt:off
        workflow.connect([
            (flt_bbr, invt_bbr, [('out_matrix_file', 'in_file')]),
            (flt_bbr, fsl2itk_fwd, [('out_matrix_file', 'transform_file')]),
        ])
        # fmt:on
        outputnode.inputs.fallback = False

        return workflow

    # Reached only if use_bbr is None
    transforms = pe.Node(niu.Merge(2), run_without_submitting=True, name='transforms')

    compare_transforms = pe.Node(niu.Function(function=compare_xforms), name='compare_transforms')

    select_transform = pe.Node(niu.Select(), run_without_submitting=True, name='select_transform')

    fsl_to_lta = pe.MapNode(LTAConvert(out_lta=True), iterfield=['in_fsl'], name='fsl_to_lta')
    # fmt:off
    workflow.connect([
        (flt_bbr, transforms, [('out_matrix_file', 'in1')]),
        (lta_to_fsl, transforms, [('out_fsl', 'in2')]),
        # Convert FSL transforms to LTA (RAS2RAS) transforms and compare
        (inputnode, fsl_to_lta, [('in_file', 'source_file')]),
        (mask_anat_brain, fsl_to_lta, [('out_file', 'target_file')]),
        (transforms, fsl_to_lta, [('out', 'in_fsl')]),
        (fsl_to_lta, compare_transforms, [('out_lta', 'lta_list')]),
        (compare_transforms, outputnode, [('out', 'fallback')]),
        # Select output transform
        (transforms, select_transform, [('out', 'inlist')]),
        (compare_transforms, select_transform, [('out', 'index')]),
        (select_transform, invt_bbr, [('out', 'in_file')]),
        (select_transform, fsl2itk_fwd, [('out', 'transform_file')]),
    ])
    # fmt:on

    return workflow


def compare_xforms(lta_list, norm_threshold=15):
    """
    Computes a normalized displacement between two affine transforms as the
    maximum overall displacement of the midpoints of the faces of a cube, when
    each transform is applied to the cube.
    This combines displacement resulting from scaling, translation and rotation.

    Although the norm is in mm, in a scaling context, it is not necessarily
    equivalent to that distance in translation.

    We choose a default threshold of 15mm as a rough heuristic.
    Normalized displacement above 20mm showed clear signs of distortion, while
    "good" BBR refinements were frequently below 10mm displaced from the rigid
    transform.
    The 10-20mm range was more ambiguous, and 15mm chosen as a compromise.
    This is open to revisiting in either direction.

    See discussion in
    `GitHub issue #681`_ <https://github.com/nipreps/fmriprep/issues/681>`_
    and the `underlying implementation
    <https://github.com/nipy/nipype/blob/56b7c81eedeeae884ba47c80096a5f66bd9f8116/nipype/algorithms/rapidart.py#L108-L159>`_.

    Parameters
    ----------

      lta_list : :obj:`list` or :obj:`tuple` of :obj:`str`
          the two given affines in LTA format
      norm_threshold : :obj:`float`
          the upper bound limit to the normalized displacement caused by the
          second transform relative to the first (default: `15`)

    """
    import nitransforms as nt
    from nipype.algorithms.rapidart import _calc_norm_affine

    bbr_affine = nt.linear.load(lta_list[0]).matrix
    fallback_affine = nt.linear.load(lta_list[1]).matrix

    norm, _ = _calc_norm_affine([fallback_affine, bbr_affine], use_differences=True)

    return norm[1] > norm_threshold


def _conditional_downsampling(in_file, in_mask, zoom_th=4.0):
    """Downsamples the input dataset for sloppy mode."""
    from pathlib import Path

    import nibabel as nb
    import nitransforms as nt
    import numpy as np
    from scipy.ndimage.filters import gaussian_filter

    img = nb.load(in_file)

    zooms = np.array(img.header.get_zooms()[:3])
    if not np.any(zooms < zoom_th):
        return in_file, in_mask

    out_file = Path('desc-resampled_input.nii.gz').absolute()
    out_mask = Path('desc-resampled_mask.nii.gz').absolute()

    shape = np.array(img.shape[:3])
    scaling = zoom_th / zooms
    newrot = np.diag(scaling).dot(img.affine[:3, :3])
    newshape = np.ceil(shape / scaling).astype(int)
    old_center = img.affine.dot(np.hstack((0.5 * (shape - 1), 1.0)))[:3]
    offset = old_center - newrot.dot((newshape - 1) * 0.5)
    newaffine = nb.affines.from_matvec(newrot, offset)

    identity = nt.Affine()

    newref = nb.Nifti1Image(np.zeros(newshape, dtype=np.uint8), newaffine)
    nt.apply(identity, img, reference=newref).to_filename(out_file)

    mask = nb.load(in_mask)
    mask.set_data_dtype(float)
    mdata = gaussian_filter(mask.get_fdata(dtype=float), scaling)
    floatmask = nb.Nifti1Image(mdata, mask.affine, mask.header)
    newmask = nt.apply(identity, floatmask, reference=newref)
    hdr = newmask.header.copy()
    hdr.set_data_dtype(np.uint8)
    newmaskdata = (newmask.get_fdata(dtype=float) > 0.5).astype(np.uint8)
    nb.Nifti1Image(newmaskdata, newmask.affine, hdr).to_filename(out_mask)

    return str(out_file), str(out_mask)



import sys
import typing as ty
from pathlib import Path

from nipype.interfaces import ants, fsl
from nipype.interfaces import utility as niu
from nipype.interfaces.ants.segmentation import JointFusion
from nipype.pipeline import engine as pe
from niworkflows.data import load as load_nwf
from niworkflows.engine import Workflow, tag
from niworkflows.interfaces.fixes import FixHeaderApplyTransforms as ApplyTransforms
from niworkflows.interfaces.fixes import FixHeaderRegistration as Registration
from niworkflows.utils.connections import listify
from smriprep.utils.misc import apply_lut
from smriprep.workflows.anatomical import (
    _aseg_to_three,
    _probseg_fast2bids,
    _split_segments,
)

from nibabies import config

LOGGER = config.loggers.workflow


@tag('anat.segmentation')
def init_segmentation_wf(
    *,
    sloppy: bool,
    method: ty.Literal['fast', 'jlf'] = 'fast',
    image_type: ty.Literal['T1w', 'T2w'] = 'T2w',
    jlf_template_dir: Path | None = None,
    omp_nthreads: int = 1,
    has_aseg: bool = False,
    name: str = 'segmentation_wf',
):
    workflow = Workflow(name=name)
    inputnode = pe.Node(
        niu.IdentityInterface(fields=['anat_brain', 'anat_aseg']),
        name='inputnode',
    )
    outputnode = pe.Node(
        niu.IdentityInterface(fields=['anat_dseg', 'anat_tpms', 'anat_aseg']),
        name='outputnode',
    )
    aseg_buffer = pe.Node(niu.IdentityInterface(fields=['anat_aseg']), name='aseg_buffer')

    to_dseg = pe.Node(
        niu.Function(function=apply_lut, output_names=['out_dseg']),
        name='to_dseg',
    )

    if has_aseg:
        LOGGER.info('ANAT Segmentation: Using existing segmentation')
        workflow.connect(inputnode, 'anat_aseg', aseg_buffer, 'anat_aseg')

    elif method == 'fast':
        workflow.__desc__ = (
            'Brain tissue segmentation of cerebrospinal fluid (CSF), white-matter (WM), and '
            f'gray-matter (GM) was performed on the brain-extracted {image_type} using FSL '
            f'FAST, distributed with {fsl.Info.version() or "version unknown"}'
        )
        # From FAST docs:
        # Number of classes to be segmented.
        # Normally you will want 3 (Grey Matter, White Matter and CSF).
        # However, if there is very poor grey/white contrast you may want to reduce this to 2;
        # alternatively, if there are strong lesions showing up as a fourth class, you may want
        # to increase this. Also, if you are segmenting T2-weighted images, you may need to
        # select 4 classes so that dark non-brain matter is processed correctly
        # (this is not a problem with T1-weighted as CSF and dark non-brain matter look similar).

        # fast_img_type = 1 if image_type == 'T1w' else 2
        # fast_n_classes = 3
        fast = pe.Node(
            fsl.FAST(
                segments=True,
                no_bias=True,
                probability_maps=True,
                # img_type=1 if image_type == 'T1w' else 2,
                # number_classes=fast_n_classes,
            ),
            name='fast',
            mem_gb=3,
        )

        to_dseg.inputs.lut = (0, 3, 1, 2)  # Maps: 0 -> 0, 3 -> 1, 1 -> 2, 2 -> 3.
        fast2bids = pe.Node(
            niu.Function(function=_probseg_fast2bids),
            name='fast2bids',
            run_without_submitting=True,
        )
        workflow.connect([
            (inputnode, fast, [('anat_brain', 'in_files')]),
            (fast, to_dseg, [('partial_volume_map', 'in_dseg')]),
            (to_dseg, outputnode, [('out_dseg', 'anat_dseg')]),
            (fast, fast2bids, [('partial_volume_files', 'inlist')]),
            (fast2bids, outputnode, [('out', 'anat_tpms')]),
        ])  # fmt:skip
        return workflow  # NOTE: no aseg will be output

    elif method == 'jlf':
        if not jlf_template_dir or not Path(jlf_template_dir).exists():
            raise RuntimeError('JLF requires a template directory.')

        jlf_wf = init_jlf_wf(
            jlf_template_dir=jlf_template_dir,
            sloppy=sloppy,
            image_type=image_type,
            omp_nthreads=omp_nthreads,
        )

        workflow.connect([
            (inputnode, jlf_wf, [('inputnode.anat_brain', 'inputnode.anat_brain')]),
            (jlf_wf, aseg_buffer, [('outputnode.anat_aseg', 'anat_aseg')]),
        ])  # fmt:skip

    to_dseg.inputs.lut = _aseg_to_three()
    split_seg = pe.Node(
        niu.Function(function=_split_segments, output_names=['out_tpms']),
        name='split_seg',
    )

    workflow.connect([
        (aseg_buffer, outputnode, [('anat_aseg', 'anat_aseg')]),
        (aseg_buffer, to_dseg, [('anat_aseg', 'in_dseg')]),
        (to_dseg, outputnode, [('out_dseg', 'anat_dseg')]),
        (to_dseg, split_seg, [('out_dseg', 'in_file')]),
        (split_seg, outputnode, [('out_tpms', 'anat_tpms')]),
    ])  # fmt:skip
    return workflow


@tag('anat.segmentation.jlf')
def init_jlf_wf(
    jlf_template_dir: Path,
    sloppy: bool,
    image_type: ty.Literal['T1w', 'T2w'] = 'T2w',
    omp_nthreads: int = 1,
    max_templates: int | None = None,
    name: str = 'jlf_wf',
):
    workflow = Workflow(name=name)
    inputnode = pe.Node(niu.IdentityInterface(fields=['anat_brain']), name='inputnode')
    outputnode = pe.Node(niu.IdentityInterface(fields=['anat_aseg']))

    jlf_templates = _parse_jlf_templates(
        jlf_template_dir,
        image_type=image_type,
        max_templates=max_templates,
    )
    segmentations = jlf_templates.keys()
    references = jlf_templates.values()

    workflow.__desc__ = (
        f'The {image_type} image was registered to {len(segmentations)} templates for '
        f'JointFusion, distributed with ANTs {ants.base.Info.version() or "version unknown"}, '
        'for image segmentation. Brain tissue segmentation of cerebrospinal fluid (CSF), '
        'white-matter (WM), and gray-matter (GM) were derived from the label fused image.'
    )

    precision = 'testing' if sloppy else 'precise'
    norm_templates = pe.MapNode(
        Registration(from_file=load_nwf(f'antsBrainExtraction_{precision}.json')),
        name='norm_templates',
        iterfield=['moving_image'],
        n_procs=omp_nthreads,
        mem_gb=config.DEFAULT_MEMORY_MIN_GB,
    )
    norm_templates.inputs.moving_image = references
    norm_templates.inputs.float = True

    apply_template = pe.MapNode(
        ApplyTransforms(
            dimension=3,
            interpolation='NearestNeighbor',
            float=True,
        ),
        iterfield=['transform', 'input_image'],
        name='apply_template',
    )
    apply_template.inputs.input_image = norm_templates

    apply_seg = pe.MapNode(
        ApplyTransforms(dimension=3, interpolation='MultiLabel'),
        name='apply_seg',
        iterfield=['transforms', 'input_image'],
    )
    apply_seg.inputs.input_image = segmentations

    jointfusion = pe.Node(
        JointFusion(
            dimension=3,
            out_label_fusion='fusion_labels.nii.gz',
            num_threads=omp_nthreads,
        ),
        name='jointfusion',
    )
    clean_label_file = pe.Node(
        niu.Function(function=_to_dtype, output_names=['out_file']), name='clean_label_file'
    )
    workflow.connect([
        (inputnode, norm_templates, [('anat_brain', 'fixed_image')]),
        (norm_templates, apply_template, [('forward_transforms', 'transforms')]),
        (inputnode, apply_template, [('anat_brain', 'reference_image')]),
        (norm_templates, apply_seg, [('forward_transforms', 'transforms')]),
        (inputnode, apply_seg, [('anat_brain', 'reference_image')]),
        (inputnode, jointfusion, [(('anat_brain', listify), 'target_image')]),
        (apply_template, jointfusion, [('output_image', 'atlas_image')]),
        (apply_seg, jointfusion, [('output_image', 'atlas_segmentation_image')]),
        (jointfusion, clean_label_file, [('out_label_fusion', 'in_file')]),
        (clean_label_file, outputnode, [('out_file', 'anat_aseg')]),
    ])  # fmt:skip
    return workflow


def _parse_jlf_templates(
    templates_dir: Path | str,
    image_type: ty.Literal['T1w', 'T2w'] = 'T2w',
    max_templates: int | None = None,
):
    """
    Parse segmentation templates directory for anatomical and segmentation files.
    The segmentations are expected to follow the FreeSurfer LUT, and the anatomicals
    should be masked.

    This is compatible with the DCAN layout::

    jlf-templates/
    ├── Template01
    │   ├── Segmentation.nii.gz
    │   ├── T1w_brain.nii.gz
    │   └── T2w_brain.nii.gz
    ├── Template02
    ...

    And the BIDS layout::

    Templates/
    ├── dataset_description.json
    ├── sub-01
    │   ├── sub-01_desc-aseg_dseg.nii.gz
    │   ├── sub-01_T1w.json
    │   ├── sub-01_T1w.nii.gz
    │   ├── sub-01_T2w.json
    │   └── sub-01_T2w.nii.gz
    ├── sub-02
    ...

    """
    segmentations = {}
    templates_dir = Path(templates_dir)
    templates = [template.name for template in templates_dir.iterdir() if template.is_dir()]
    if not max_templates:
        max_templates = len(templates)

    if not templates:
        raise FileNotFoundError('JLF requested but no templates found.')

    for template in templates[:max_templates]:
        files = sorted((templates_dir / template).iterdir())
        seg = None
        anat = None
        for fl in files:
            if 'Segmentation' in fl.name or '_dseg' in fl.name:
                seg = str(fl)
            elif image_type in fl.name:
                anat = str(fl)
        if seg is None or anat is None:
            print(
                f'No anatomical or segmentation found for JLF template: {template}',
                file=sys.stderr,
            )
            continue
        segmentations[seg] = anat

    if len(segmentations) == 0:
        raise FileNotFoundError('JLF requested but anatomicals / segmentations were not found.')
    return segmentations


def _to_dtype(in_file, dtype='uint8'):
    """
    Freesurfer's ``mri_convert`` complains about unsigned 32-bit integers.
    Since we may use the JLF segmentation with FreeSurfer tools
    better to make this change now.
    """
    from pathlib import Path

    import nibabel as nb
    import numpy as np

    img = nb.load(in_file)
    out_file = Path(f'labels{"".join(Path(in_file).suffixes)}').absolute()

    new_data = np.asanyarray(img.get_fdata(), dtype=dtype)
    img.set_data_dtype(dtype)
    img.__class__(new_data, img.affine, img.header).to_filename(out_file)
    return str(out_file)



import shutil
from pathlib import Path

import pytest

from nibabies.interfaces.mcribs import MCRIBReconAll

SUBJECT_ID = 'X'


@pytest.fixture
def mcribs_directory(tmp_path):
    def make_tree(path, tree):
        for d, fls in tree.items():
            (path / d).mkdir(exist_ok=True)
            for f in fls:
                (path / d / f).touch()

    root = tmp_path / 'mcribs'
    surfrecon = root / SUBJECT_ID / 'SurfReconDeformable' / SUBJECT_ID
    surfrecon.mkdir(parents=True, exist_ok=True)
    make_tree(surfrecon, MCRIBReconAll._expected_files['surfrecon'])
    autorecon = root / SUBJECT_ID / 'freesurfer' / SUBJECT_ID
    autorecon.mkdir(parents=True, exist_ok=True)
    make_tree(autorecon, MCRIBReconAll._expected_files['autorecon'])

    yield root

    shutil.rmtree(root)


def test_MCRIBReconAll(mcribs_directory):
    t2w = Path('T2w.nii.gz')
    t2w.touch()

    surfrecon = MCRIBReconAll(
        subject_id=SUBJECT_ID,
        surfrecon=True,
        surfrecon_method='Deformable',
        join_thresh=1.0,
        fast_collision=True,
    )

    # Requires T2w input
    with pytest.raises(AttributeError):
        surfrecon.cmdline  # noqa

    surfrecon.inputs.t2w_file = t2w
    # Since no existing directory is found, will run fresh
    assert 'MCRIBReconAll --deformablefastcollision --deformablejointhresh' in surfrecon.cmdline

    # But should not need to run again
    surfrecon.inputs.outdir = mcribs_directory
    assert surfrecon.cmdline == 'echo MCRIBReconAll: nothing to do'

    t2w.unlink()

    autorecon = MCRIBReconAll(
        subject_id=SUBJECT_ID,
        autorecon_after_surf=True,
    )
    # No need for T2w here
    assert autorecon.cmdline == 'MCRIBReconAll --autoreconaftersurf X'
    autorecon.inputs.outdir = mcribs_directory
    assert autorecon.cmdline == 'echo MCRIBReconAll: nothing to do'



#!/usr/bin/env python3
"""Update and sort the creators list of the zenodo record."""
import sys
from pathlib import Path
import json
from fuzzywuzzy import fuzz, process

# These ORCIDs should go last
CREATORS_LAST = ['Poldrack, Russell A.', 'Fair, Damien A.']
CONTRIBUTORS_LAST = []


def sort_contributors(entries, git_lines, exclude=None, last=None):
    """Return a list of author dictionaries, ordered by contribution."""
    last = last or []
    sorted_authors = sorted(entries, key=lambda i: i['name'])

    first_last = [' '.join(val['name'].split(',')[::-1]).strip()
                  for val in sorted_authors]
    first_last_excl = [' '.join(val['name'].split(',')[::-1]).strip()
                       for val in exclude or []]

    unmatched = []
    author_matches = []
    position = 1
    for ele in git_lines:
        matches = process.extract(ele, first_last, scorer=fuzz.token_sort_ratio,
                                  limit=2)
        if not matches:
            return [], []
        # matches is a list [('First match', % Match), ('Second match', % Match)]
        if matches[0][1] > 80:
            val = sorted_authors[first_last.index(matches[0][0])]
        else:
            # skip unmatched names
            if ele not in first_last_excl:
                unmatched.append(ele)
            continue

        if val not in author_matches:
            val['position'] = position
            author_matches.append(val)
            position += 1

    names = {' '.join(val['name'].split(',')[::-1]).strip() for val in author_matches}
    for missing_name in first_last:
        if missing_name not in names:
            missing = sorted_authors[first_last.index(missing_name)]
            missing['position'] = position
            author_matches.append(missing)
            position += 1

    all_names = [val['name'] for val in author_matches]
    for last_author in last:
        author_matches[all_names.index(last_author)]['position'] = position
        position += 1

    author_matches = sorted(author_matches, key=lambda k: k['position'])

    return author_matches, unmatched


def get_git_lines(fname='line-contributors.txt'):
    """Run git-line-summary."""
    import shutil
    import subprocess as sp
    contrib_file = Path(fname)

    lines = []
    if contrib_file.exists():
        print('WARNING: Reusing existing line-contributors.txt file.', file=sys.stderr)
        lines = contrib_file.read_text().splitlines()

    cmd = [shutil.which('git-line-summary')]
    if cmd == [None]:
        cmd = [shutil.which('git-summary'), "--line"]
    if not lines and cmd[0]:
        print(f"Running {' '.join(cmd)!r} on repo")
        lines = sp.check_output(cmd).decode().splitlines()
        lines = [l for l in lines if "Not Committed Yet" not in l]
        contrib_file.write_text('\n'.join(lines))

    if not lines:
        raise RuntimeError("""\
Could not find line-contributors from git repository.%s""" % """ \
git-(line-)summary not found, please install git-extras. """ * (cmd[0] is None))
    return [' '.join(line.strip().split()[1:-1]) for line in lines if '%' in line]


if __name__ == '__main__':
    data = get_git_lines()

    zenodo_file = Path('.zenodo.json')
    zenodo = json.loads(zenodo_file.read_text())

    creators = json.loads(Path('.maint/developers.json').read_text())
    zen_creators, miss_creators = sort_contributors(
        creators, data,
        exclude=json.loads(Path('.maint/former.json').read_text()),
        last=CREATORS_LAST)
    contributors = json.loads(Path('.maint/contributors.json').read_text())
    zen_contributors, miss_contributors = sort_contributors(
        contributors, data,
        exclude=json.loads(Path('.maint/former.json').read_text()),
        last=CONTRIBUTORS_LAST)
    zenodo['creators'] = zen_creators
    zenodo['contributors'] = zen_contributors

    print("Some people made commits, but are missing in .maint/ "
          "files: %s." % ', '.join(set(miss_creators).intersection(miss_contributors)),
          file=sys.stderr)

    # Remove position
    for creator in zenodo['creators']:
        del creator['position']
        if isinstance(creator['affiliation'], list):
            creator['affiliation'] = creator['affiliation'][0]

    for creator in zenodo['contributors']:
        creator['type'] = 'Researcher'
        del creator['position']
        if isinstance(creator['affiliation'], list):
            creator['affiliation'] = creator['affiliation'][0]

    zenodo_file.write_text('%s\n' % json.dumps(zenodo, indent=2, ensure_ascii=False))



import json
import os
from pathlib import Path

import pooch
from nipype.interfaces.base import (
    DynamicTraitedSpec,
    File,
    SimpleInterface,
    TraitedSpec,
    traits,
)

import nibabies


class _RetrievePoochFilesInputSpec(DynamicTraitedSpec):
    intermediate = traits.Str(required=True, desc='the intermediate space')
    target = traits.Str(required=True, desc='the target space')


class _RetrievePoochFilesOutputSpec(TraitedSpec):
    int2tgt_xfm = File(desc='Intermediate to target transform')
    tgt2int_xfm = File(desc='Target to intermediate transform')


class RetrievePoochFiles(SimpleInterface):
    input_spec = _RetrievePoochFilesInputSpec
    output_spec = _RetrievePoochFilesOutputSpec

    def _run_interface(self, runtime):
        int2tgt, tgt2int = _retrieve_xfms(self.inputs.intermediate, self.inputs.target)
        self._results['int2tgt_xfm'] = int2tgt
        self._results['tgt2int_xfm'] = tgt2int
        return runtime


def _retrieve_xfms(
    intermediate: str,
    target: str,
):
    """Fetch transforms from the OSF repository (https://osf.io/y763j/)."""

    manifest = json.loads(nibabies.data.load('xfm_manifest.json').read_text())

    def sanitize(space):
        # MNIInfant:cohort-1 -> MNIInfant+1
        return space.replace(':cohort-', '+')

    intmd = sanitize(intermediate)
    tgt = sanitize(target)

    cache_dir = Path(os.getenv('NIBABIES_POOCH_DIR', Path.cwd()))

    int2std_name = f'from-{intmd}_to-{tgt}_xfm.h5'
    int2std_meta = manifest[int2std_name]
    int2std = pooch.retrieve(
        url=int2std_meta['url'],
        path=cache_dir,
        known_hash=int2std_meta['hash'],
        fname=int2std_name,
    )

    std2int_name = f'from-{tgt}_to-{intmd}_xfm.h5'
    std2int_meta = manifest[std2int_name]
    std2int = pooch.retrieve(
        url=std2int_meta['url'],
        path=cache_dir,
        known_hash=std2int_meta['hash'],
        fname=std2int_name,
    )

    return int2std, std2int



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
#
# Copyright 2023 The NiPreps Developers <nipreps@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# We support and encourage derived works from this project, please read
# about our expectations at
#
#     https://www.nipreps.org/community/licensing/
#
"""
Multi-echo EPI
~~~~~~~~~~~~~~

For using multi-echo EPI data.

Change directory to provide relative paths for doctests
>>> import os
>>> filepath = os.path.dirname( os.path.realpath( __file__ ) )
>>> datadir = os.path.realpath(os.path.join(filepath, '../data/'))
>>> os.chdir(datadir)

"""

import os

from nipype import logging
from nipype.interfaces.base import (
    CommandLine,
    CommandLineInputSpec,
    File,
    TraitedSpec,
    traits,
)

LOGGER = logging.getLogger('nipype.interface')


class T2SMapInputSpec(CommandLineInputSpec):
    in_files = traits.List(
        File(exists=True),
        argstr='-d %s',
        position=1,
        mandatory=True,
        minlen=3,
        desc='multi-echo BOLD EPIs',
    )
    echo_times = traits.List(
        traits.Float,
        argstr='-e %s',
        position=2,
        mandatory=True,
        minlen=3,
        desc='echo times',
    )
    mask_file = File(
        argstr='--mask %s',
        position=3,
        desc='mask file',
        exists=True,
    )
    fittype = traits.Enum(
        'curvefit',
        'loglin',
        argstr='--fittype %s',
        position=4,
        usedefault=True,
        desc='Desired fitting method: '
        '"loglin" means that a linear model is fit '
        'to the log of the data. '
        '"curvefit" means that a more computationally '
        'demanding monoexponential model is fit '
        'to the raw data.',
    )


class T2SMapOutputSpec(TraitedSpec):
    t2star_map = File(exists=True, desc='limited T2* map')
    s0_map = File(exists=True, desc='limited S0 map')
    optimal_comb = File(exists=True, desc='optimally combined ME-EPI time series')


class T2SMap(CommandLine):
    """
    Runs the tedana T2* workflow to generate an adaptive T2* map and create
    an optimally combined ME-EPI time series.

    Example
    =======
    >>> from nibabies.interfaces import multiecho
    >>> t2smap = multiecho.T2SMap()
    >>> t2smap.inputs.in_files = ['sub-01_run-01_echo-1_bold.nii.gz', \
                                  'sub-01_run-01_echo-2_bold.nii.gz', \
                                  'sub-01_run-01_echo-3_bold.nii.gz']
    >>> t2smap.inputs.echo_times = [0.013, 0.027, 0.043]
    >>> t2smap.cmdline  # doctest: +ELLIPSIS
    't2smap -d sub-01_run-01_echo-1_bold.nii.gz sub-01_run-01_echo-2_bold.nii.gz \
sub-01_run-01_echo-3_bold.nii.gz -e 13.0 27.0 43.0 --fittype curvefit'

    """

    _cmd = 't2smap'
    input_spec = T2SMapInputSpec
    output_spec = T2SMapOutputSpec

    def _format_arg(self, name, trait_spec, value):
        if name == 'echo_times':
            value = [te * 1000 for te in value]
        return super()._format_arg(name, trait_spec, value)

    def _list_outputs(self):
        outputs = self._outputs().get()
        out_dir = os.getcwd()
        outputs['t2star_map'] = os.path.join(out_dir, 'T2starmap.nii.gz')
        outputs['s0_map'] = os.path.join(out_dir, 'S0map.nii.gz')
        outputs['optimal_comb'] = os.path.join(out_dir, 'desc-optcom_bold.nii.gz')
        return outputs



#!/usr/bin/env python

'Preemptive caching of commonly used TemplateFlow templates'
import templateflow.api as tf


def fetch_MNI6():
    """
    Expected templates:

    tpl-MNI152NLin6Asym/tpl-MNI152NLin6Asym_res-01_T1w.nii.gz
    tpl-MNI152NLin6Asym/tpl-MNI152NLin6Asym_res-02_T1w.nii.gz
    tpl-MNI152NLin6Asym/tpl-MNI152NLin6Asym_res-01_desc-brain_mask.nii.gz
    tpl-MNI152NLin6Asym/tpl-MNI152NLin6Asym_res-02_desc-brain_mask.nii.gz
    tpl-MNI152NLin6Asym/tpl-MNI152NLin6Asym_res-02_atlas-HCP_dseg.nii.gz
    """
    template = 'MNI152NLin6Asym'

    tf.get(template, resolution=(1, 2), desc=None, suffix='T1w')
    tf.get(template, resolution=(1, 2), desc='brain', suffix='mask')
    # CIFTI
    tf.get(template, resolution=2, atlas='HCP', suffix='dseg')


def fetch_UNCInfant():
    """
    Expected templates:

    tpl-UNCInfant/cohort-1/tpl-UNCInfant_cohort-1_T1w.nii.gz
    tpl-UNCInfant/cohort-1/tpl-UNCInfant_cohort-1_label-brain_probseg.nii.gz
    tpl-UNCInfant/cohort-1/tpl-UNCInfant_cohort-1_label-brain_mask.nii.gz
    tpl-UNCInfant/cohort-1/tpl-UNCInfant_cohort-1_label-BrainCerebellumExtraction_mask.nii.gz
    """
    template = 'UNCInfant'

    tf.get(template, cohort=1, desc=None, suffix='T1w')
    tf.get(template, cohort=1, label='brain', suffix='probseg')
    tf.get(template, cohort=1, label='brain', suffix='mask')
    tf.get(template, cohort=1, label='BrainCerebellumExtraction', suffix='mask')


def fetch_fsaverage():
    """
    Expected templates:

    tpl-fsaverage/tpl-fsaverage_hemi-L_den-164k_desc-std_sphere.surf.gii
    tpl-fsaverage/tpl-fsaverage_hemi-R_den-164k_desc-std_sphere.surf.gii
    tpl-fsaverage/tpl-fsaverage_hemi-L_den-164k_desc-vaavg_midthickness.shape.gii
    tpl-fsaverage/tpl-fsaverage_hemi-R_den-164k_desc-vaavg_midthickness.shape.gii
    tpl-fsaverage/tpl-fsaverage_hemi-L_den-41k_sphere.surf.gii
    tpl-fsaverage/tpl-fsaverage_hemi-R_den-41k_sphere.surf.gii
    """
    template = 'fsaverage'

    tf.get(template, density='164k', desc='std', suffix='sphere')
    tf.get(template, density='164k', desc='vaavg', suffix='midthickness')
    tf.get(template, density='41k', desc=None, suffix='sphere', extension='.surf.gii')


def fetch_fsLR():
    """
    Expected templates:

    tpl-fsLR/tpl-fsLR_hemi-L_den-32k_desc-nomedialwall_dparc.label.gii
    tpl-fsLR/tpl-fsLR_hemi-L_den-32k_desc-vaavg_midthickness.shape.gii
    tpl-fsLR/tpl-fsLR_hemi-L_den-32k_sphere.surf.gii
    tpl-fsLR/tpl-fsLR_hemi-R_den-32k_desc-nomedialwall_dparc.label.gii
    tpl-fsLR/tpl-fsLR_hemi-R_den-32k_desc-vaavg_midthickness.shape.gii
    tpl-fsLR/tpl-fsLR_hemi-R_den-32k_sphere.surf.gii
    tpl-fsLR/tpl-fsLR_space-fsaverage_hemi-L_den-32k_sphere.surf.gii
    tpl-fsLR/tpl-fsLR_space-fsaverage_hemi-R_den-32k_sphere.surf.gii
    """
    tf.get('fsLR', density='32k')
    tf.get('fsLR', density='164k', desc='std', suffix='sphere')
    tf.get('fsLR', density='164k', suffix='midthickness')


def fetch_MNIInfant(cohort=1):
    """
    Expected templates:

    tpl-MNIInfant/cohort-1/tpl-MNIInfant_cohort-1_res-1_T1w.nii.gz
    tpl-MNIInfant/cohort-1/tpl-MNIInfant_cohort-1_res-1_T2w.nii.gz
    tpl-MNIInfant/cohort-1/tpl-MNIInfant_cohort-1_res-1_desc-brain_mask.nii.gz
    tpl-MNIInfant/cohort-1/tpl-MNIInfant_cohort-1_res-2_T1w.nii.gz
    tpl-MNIInfant/cohort-1/tpl-MNIInfant_cohort-1_res-2_T1w.nii.gz
    tpl-MNIInfant/cohort-1/tpl-MNIInfant_cohort-1_res-2_desc-brain_mask.nii.gz
    """
    template = 'MNIInfant'

    tf.get(template, cohort=cohort, suffix='T1w')
    tf.get(template, cohort=cohort, suffix='T2w')
    tf.get(template, cohort=cohort, desc='brain', suffix='mask')


def fetch_dhcpAsym(cohort=42):
    """
    Expected templates:

    tpl-dhcpAsym_cohort-42_hemi-L_den-32k_sphere.surf.gii
    tpl-dhcpAsym_cohort-42_hemi-R_den-32k_sphere.surf.gii
    tpl-dhcpAsym_cohort-42_space-fsaverage_hemi-L_den-41k_desc-reg_sphere.surf.gii
    tpl-dhcpAsym_cohort-42_space-fsaverage_hemi-R_den-41k_desc-reg_sphere.surf.gii
    """
    template = 'dhcpAsym'

    tf.get(template, cohort=cohort, density='32k', desc=None, suffix='sphere')
    tf.get(template, cohort=cohort, space='fsaverage', density='41k', desc='reg', suffix='sphere')


def fetch_MNI2009():
    template = 'MNI152NLin2009cAsym'

    tf.get(template, resolution=(1, 2), desc=None, suffix='T1w')
    tf.get(template, resolution=(1, 2), desc='brain', suffix='mask')
    tf.get(template, resolution=2, desc='fMRIPrep', suffix='boldref')
    tf.get(template, resolution=1, label='brain', suffix='probseg')


def main():
    fetch_MNI6()
    fetch_UNCInfant()
    fetch_fsaverage()
    fetch_fsLR()
    fetch_MNIInfant()
    fetch_dhcpAsym()
    fetch_MNI2009()


if __name__ == '__main__':
    main()



"""Utilities for loading transforms for resampling"""

from pathlib import Path

import nitransforms as nt


def load_transforms(xfm_paths: list[Path], inverse: list[bool]) -> nt.base.TransformBase:
    """Load a series of transforms as a nitransforms TransformChain

    An empty list will return an identity transform
    """
    if len(inverse) == 1:
        inverse *= len(xfm_paths)
    elif len(inverse) != len(xfm_paths):
        raise ValueError('Mismatched number of transforms and inverses')

    chain = None
    for path, inv in zip(xfm_paths[::-1], inverse[::-1], strict=False):
        path = Path(path)
        if path.suffix == '.h5':
            # Load as a TransformChain
            xfm = nt.manip.load(path)
        else:
            xfm = nt.linear.load(path)
        if inv:
            xfm = ~xfm
        if chain is None:
            chain = xfm
        else:
            chain += xfm
    if chain is None:
        chain = nt.Affine()  # Identity
    return chain



import nipype.interfaces.utility as niu
import nipype.pipeline.engine as pe
from niworkflows.engine import Workflow, tag


@tag('anat.preproc')
def init_anat_preproc_wf(
    *,
    bspline_fitting_distance: int = 200,
    name: str = 'anat_preproc_wf',
) -> Workflow:
    """Polish up raw anatomical data.

    This workflow accepts T1w/T2w images as inputs (either raw or a merged template) and performs:
    - Intensity clipping
    - N4 Bias Field Correction

    The outputs of this workflow will be a structural reference used for subsequent processing.

    Inputs
    ------
    in_anat : :obj:`str`
        A single volume T1w/T2w image

    Outputs
    -------
    anat_preproc: :obj:`str`
        Preprocessed anatomical image (Denoising/INU/Clipping)
    """
    from nipype.interfaces.ants import N4BiasFieldCorrection
    from niworkflows.interfaces.header import ValidateImage
    from niworkflows.interfaces.nibabel import IntensityClip

    wf = Workflow(name=name)
    inputnode = pe.Node(
        niu.IdentityInterface(fields=['in_anat']),
        name='inputnode',
    )
    outputnode = pe.Node(
        niu.IdentityInterface(fields=['anat_preproc']),
        name='outputnode',
    )

    # validate image
    validate = pe.Node(ValidateImage(), name='anat_validate', run_without_submitting=True)
    clip = pe.Node(IntensityClip(p_min=10.0, p_max=99.5), name='clip')
    n4_correct = pe.Node(
        N4BiasFieldCorrection(
            dimension=3,
            bspline_fitting_distance=bspline_fitting_distance,
            save_bias=True,
            copy_header=True,
            n_iterations=[50] * 5,
            convergence_threshold=1e-7,
            rescale_intensities=True,
            shrink_factor=4,
        ),
        name='n4_correct',
    )
    final_clip = pe.Node(IntensityClip(p_min=5.0, p_max=99.5), name='final_clip')

    wf.connect([
        (inputnode, validate, [('in_anat', 'in_file')]),
        (validate, clip, [('out_file', 'in_file')]),
        (clip, n4_correct, [('out_file', 'input_image')]),
        (n4_correct, final_clip, [('output_image', 'in_file')]),
        (final_clip, outputnode, [('out_file', 'anat_preproc')]),
    ])  # fmt:skip
    return wf


@tag('anat.csf_norm')
def init_csf_norm_wf(name: str = 'csf_norm_wf') -> Workflow:
    """Replace low intensity voxels within the CSF mask with the median value."""

    workflow = Workflow(name=name)
    workflow.__desc__ = (
        'The CSF mask was used to normalize the anatomical template by the median of voxels '
        'within the mask.'
    )
    inputnode = pe.Node(
        niu.IdentityInterface(fields=['anat_preproc', 'anat_tpms']),
        name='inputnode',
    )
    outputnode = pe.Node(niu.IdentityInterface(fields=['anat_preproc']), name='outputnode')

    # select CSF from BIDS-ordered list (GM, WM, CSF)
    select_csf = pe.Node(niu.Select(index=2), name='select_csf')
    norm_csf = pe.Node(niu.Function(function=_normalize_roi), name='norm_csf')

    workflow.connect([
        (inputnode, select_csf, [('anat_tpms', 'inlist')]),
        (select_csf, norm_csf, [('out', 'mask_file')]),
        (inputnode, norm_csf, [('anat_preproc', 'in_file')]),
        (norm_csf, outputnode, [('out', 'anat_preproc')]),
    ])  # fmt:skip

    return workflow


def init_conform_derivative_wf(
    *, in_file: str = None, name: str = 'conform_derivative_wf'
) -> pe.Workflow:
    """
    Ensure derivatives share the same space as anatomical references.

    This workflow is used when a derivative is provided without a reference.
    """
    from niworkflows.interfaces.header import MatchHeader
    from niworkflows.interfaces.images import Conform, TemplateDimensions

    workflow = pe.Workflow(name=name)
    inputnode = pe.Node(niu.IdentityInterface(fields=['in_file', 'ref_file']), name='inputnode')
    inputnode.inputs.in_file = in_file
    outputnode = pe.Node(niu.IdentityInterface(fields=['out_file']), name='outputnode')

    ref_dims = pe.Node(TemplateDimensions(), name='ref_dims')
    conform = pe.Node(Conform(), name='conform')
    # Avoid mismatch tolerance from input
    match_header = pe.Node(MatchHeader(), name='match_header')

    workflow.connect([
        (inputnode, ref_dims, [('ref_file', 'anat_list')]),
        (ref_dims, conform, [
            ('target_zooms', 'target_zooms'),
            ('target_shape', 'target_shape'),
        ]),
        (inputnode, conform, [('in_file', 'in_file')]),
        (conform, match_header, [('out_file', 'in_file')]),
        (inputnode, match_header, [('ref_file', 'reference')]),
        (match_header, outputnode, [('out_file', 'out_file')]),
    ])  # fmt:skip
    return workflow


def _normalize_roi(in_file, mask_file, threshold=0.2, out_file=None):
    """Normalize low intensity voxels that fall within a given mask."""
    import nibabel as nb
    import numpy as np

    img = nb.load(in_file)
    img_data = np.asanyarray(img.dataobj)
    mask_img = nb.load(mask_file)
    # binary mask
    bin_mask = np.asanyarray(mask_img.dataobj) > threshold
    mask_data = bin_mask * img_data
    masked_data = mask_data[mask_data > 0]

    median = np.median(masked_data).astype(masked_data.dtype)
    normed_data = np.maximum(img_data, bin_mask * median)

    oimg = img.__class__(normed_data, img.affine, img.header)
    if not out_file:
        from nipype.utils.filemanip import fname_presuffix

        out_file = fname_presuffix(in_file, suffix='normed')
    oimg.to_filename(out_file)
    return out_file



from __future__ import annotations

import json
import shutil
from collections import defaultdict
from pathlib import Path

from bids.layout import BIDSLayout
from niworkflows.data import load as nwf_load

from nibabies.data import load


def collect_anatomical_derivatives(
    derivatives_dir: Path | str,
    subject_id: str,
    std_spaces: list,
    session_id: str | None,
    spec: dict | None = None,
    patterns: list | None = None,
):
    """
    Collect outputs from across processing stages.

    Potential files:
    - T1w preproc
    - T2w preproc
    - T1w mask
    - T2w mask


    """

    if spec is None or patterns is None:
        _spec, _patterns = tuple(json.loads(load('io_spec_anat.json').read_text()).values())

        if spec is None:
            spec = _spec
        if patterns is None:
            patterns = _patterns

    deriv_config = nwf_load('nipreps.json')
    layout = BIDSLayout(derivatives_dir, config=deriv_config, validate=False)
    derivs_cache = {}

    base_qry = {
        'subject': subject_id,
    }
    if session_id is not None:
        base_qry['session'] = session_id

    for key, qry in spec['baseline'].items():
        qry.update(base_qry)
        item = layout.get(return_type='filename', **qry)
        if not item:
            continue

        derivs_cache[key] = item[0] if len(item) == 1 else item

    for key, qry in spec['coreg'].items():  # T1w->T2w, T2w->T1w
        qry.update(base_qry)
        item = layout.get(return_type='filename', **qry)
        if not item:
            continue
        derivs_cache[key] = item[0] if len(item) == 1 else item

    transforms = derivs_cache.setdefault('transforms', {})
    for _space in std_spaces:
        space = _space.replace(':cohort-', '+')
        for key, qry in spec['transforms'].items():
            qry = qry.copy()
            qry.update(base_qry)
            qry['from'] = qry['from'] or space
            qry['to'] = qry['to'] or space
            item = layout.get(return_type='filename', **qry)
            if not item:
                continue
            transforms.setdefault(_space, {})[key] = item[0] if len(item) == 1 else item

    for key, qry in spec['surfaces'].items():
        qry.update(base_qry)
        item = layout.get(return_type='filename', **qry)
        if not item or len(item) != 2:
            continue

        derivs_cache[key] = sorted(item)

    return derivs_cache


def collect_functional_derivatives(
    derivatives_dir: Path,
    entities: dict,
    fieldmap_id: str | None,
    spec: dict | None = None,
    patterns: list[str] | None = None,
):
    """Gather existing derivatives and compose a cache."""
    if spec is None or patterns is None:
        _spec, _patterns = tuple(
            json.loads(load.readable('io_spec_func.json').read_text()).values()
        )

        if spec is None:
            spec = _spec
        if patterns is None:
            patterns = _patterns

    derivs_cache = defaultdict(list, {})
    deriv_config = nwf_load('nipreps.json')
    layout = BIDSLayout(derivatives_dir, config=deriv_config, validate=False)
    derivatives_dir = Path(derivatives_dir)

    # search for both boldrefs
    for key, qry in spec['baseline'].items():
        query = {**qry, **entities}
        item = layout.get(return_type='filename', **query)
        if not item:
            continue
        derivs_cache[f'{key}_boldref'] = item[0] if len(item) == 1 else item

    for xfm, qry in spec['transforms'].items():
        query = {**qry, **entities}
        if xfm == 'boldref2fmap':
            query['to'] = fieldmap_id
        item = layout.get(return_type='filename', **query)
        if not item:
            continue
        derivs_cache[xfm] = item[0] if len(item) == 1 else item
    return derivs_cache


def copy_derivatives(
    derivs: dict,
    outdir: Path,
    modality: str,
    subject_id: str,
    session_id: str | None = None,
) -> None:
    """
    Creates a copy of any found derivatives into output directory.

    Attempts to preserve file metadata to distinguish from generated files.
    """
    out_levels = [subject_id, modality]
    if session_id:
        out_levels.insert(1, session_id)

    outpath = outdir.joinpath(*out_levels)
    outpath.mkdir(parents=True, exist_ok=True)

    for deriv in derivs.values():
        # Skip empty, lists
        if not isinstance(deriv, str):
            continue
        deriv = Path(deriv)

        shutil.copy2(deriv, outpath / deriv.name)
        json = deriv.parent / (deriv.name.split('.')[0] + '.json')
        if json.exists():
            shutil.copy2(json, outpath / json.name)



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
#
# Copyright The NiPreps Developers <nipreps@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# We support and encourage derived works from this project, please read
# about our expectations at
#
#     https://www.nipreps.org/community/licensing/
#
"""
Calculate BOLD confounds
^^^^^^^^^^^^^^^^^^^^^^^^

.. autofunction:: init_bold_confs_wf

"""

from nipype.algorithms import confounds as nac
from nipype.interfaces import utility as niu
from nipype.pipeline import engine as pe

from nibabies.config import DEFAULT_DISMISS_ENTITIES, DEFAULT_MEMORY_MIN_GB
from nibabies.interfaces import DerivativesDataSink
from nibabies.interfaces.confounds import (
    FilterDropped,
    FMRISummary,
    GatherConfounds,
    RenameACompCor,
)


def init_bold_confs_wf(
    mem_gb: float,
    metadata: dict,
    regressors_all_comps: bool,
    regressors_dvars_th: float,
    regressors_fd_th: float,
    freesurfer: bool = False,
    name: str = 'bold_confs_wf',
):
    """
    Build a workflow to generate and write out confounding signals.

    This workflow calculates confounds for a BOLD series, and aggregates them
    into a :abbr:`TSV (tab-separated value)` file, for use as nuisance
    regressors in a :abbr:`GLM (general linear model)`.
    The following confounds are calculated, with column headings in parentheses:

    #. Region-wise average signal (``csf``, ``white_matter``, ``global_signal``)
    #. DVARS - original and standardized variants (``dvars``, ``std_dvars``)
    #. Framewise displacement, based on head-motion parameters
       (``framewise_displacement``)
    #. Temporal CompCor (``t_comp_cor_XX``)
    #. Anatomical CompCor (``a_comp_cor_XX``)
    #. Cosine basis set for high-pass filtering w/ 0.008 Hz cut-off
       (``cosine_XX``)
    #. Non-steady-state volumes (``non_steady_state_XX``)
    #. Estimated head-motion parameters, in mm and rad
       (``trans_x``, ``trans_y``, ``trans_z``, ``rot_x``, ``rot_y``, ``rot_z``)


    Prior to estimating aCompCor and tCompCor, non-steady-state volumes are
    censored and high-pass filtered using a :abbr:`DCT (discrete cosine
    transform)` basis.
    The cosine basis, as well as one regressor per censored volume, are included
    for convenience.

    Workflow Graph
        .. workflow::
            :graph2use: orig
            :simple_form: yes

            from fmriprep.workflows.bold.confounds import init_bold_confs_wf
            wf = init_bold_confs_wf(
                mem_gb=1,
                metadata={},
                regressors_all_comps=False,
                regressors_dvars_th=1.5,
                regressors_fd_th=0.5,
            )

    Parameters
    ----------
    mem_gb : :obj:`float`
        Size of BOLD file in GB - please note that this size
        should be calculated after resamplings that may extend
        the FoV
    metadata : :obj:`dict`
        BIDS metadata for BOLD file
    name : :obj:`str`
        Name of workflow (default: ``bold_confs_wf``)
    regressors_all_comps : :obj:`bool`
        Indicates whether CompCor decompositions should return all
        components instead of the minimal number of components necessary
        to explain 50 percent of the variance in the decomposition mask.
    regressors_dvars_th : :obj:`float`
        Criterion for flagging DVARS outliers
    regressors_fd_th : :obj:`float`
        Criterion for flagging framewise displacement outliers

    Inputs
    ------
    bold
        BOLD image, after the prescribed corrections (STC, HMC and SDC)
        when available.
    bold_mask
        BOLD series mask
    movpar_file
        SPM-formatted motion parameters file
    rmsd_file
        Root mean squared deviation as measured by ``fsl_motion_outliers`` [Jenkinson2002]_.
    skip_vols
        number of non steady state volumes
    anat_mask
        Mask of the skull-stripped template image
    anat_tpms
        List of tissue probability maps in anatomical space
    boldref2anat_xfm
        Affine matrix that maps the BOLD reference space into alignment with
        the anatomical space

    Outputs
    -------
    confounds_file
        TSV of all aggregated confounds
    rois_report
        Reportlet visualizing white-matter/CSF mask used for aCompCor,
        the ROI for tCompCor and the BOLD brain mask.
    confounds_metadata
        Confounds metadata dictionary.
    crown_mask
        Mask of brain edge voxels

    """
    from nireports.interfaces.nuisance import (
        CompCorVariancePlot,
        ConfoundsCorrelationPlot,
    )
    from niworkflows.engine.workflows import LiterateWorkflow as Workflow
    from niworkflows.interfaces.confounds import ExpandModel, SpikeRegressors
    from niworkflows.interfaces.fixes import FixHeaderApplyTransforms as ApplyTransforms
    from niworkflows.interfaces.images import SignalExtraction
    from niworkflows.interfaces.morphology import BinaryDilation, BinarySubtraction
    from niworkflows.interfaces.nibabel import ApplyMask, Binarize
    from niworkflows.interfaces.patches import RobustACompCor as ACompCor
    from niworkflows.interfaces.patches import RobustTCompCor as TCompCor
    from niworkflows.interfaces.reportlets.masks import ROIsPlot
    from niworkflows.interfaces.utility import TSV2JSON, AddTSVHeader, DictMerge

    from ...interfaces.confounds import aCompCorMasks

    gm_desc = (
        "dilating a GM mask extracted from the FreeSurfer's *aseg* segmentation"
        if freesurfer
        else 'thresholding the corresponding partial volume map at 0.05'
    )

    workflow = Workflow(name=name)
    workflow.__desc__ = f"""\
Several confounding time-series were calculated based on the
*preprocessed BOLD*: framewise displacement (FD), DVARS and
three region-wise global signals.
FD was computed using two formulations following Power (absolute sum of
relative motions, @power_fd_dvars) and Jenkinson (relative root mean square
displacement between affines, @mcflirt).
FD and DVARS are calculated for each functional run, both using their
implementations in *Nipype* [following the definitions by @power_fd_dvars].
The three global signals are extracted within the CSF, the WM, and
the whole-brain masks.
Additionally, a set of physiological regressors were extracted to
allow for component-based noise correction [*CompCor*, @compcor].
Principal components are estimated after high-pass filtering the
*preprocessed BOLD* time-series (using a discrete cosine filter with
128s cut-off) for the two *CompCor* variants: temporal (tCompCor)
and anatomical (aCompCor).
tCompCor components are then calculated from the top 2% variable
voxels within the brain mask.
For aCompCor, three probabilistic masks (CSF, WM and combined CSF+WM)
are generated in anatomical space.
The implementation differs from that of Behzadi et al. in that instead
of eroding the masks by 2 pixels on BOLD space, a mask of pixels that
likely contain a volume fraction of GM is subtracted from the aCompCor masks.
This mask is obtained by {gm_desc}, and it ensures components are not extracted
from voxels containing a minimal fraction of GM.
Finally, these masks are resampled into BOLD space and binarized by
thresholding at 0.99 (as in the original implementation).
Components are also calculated separately within the WM and CSF masks.
For each CompCor decomposition, the *k* components with the largest singular
values are retained, such that the retained components' time series are
sufficient to explain 50 percent of variance across the nuisance mask (CSF,
WM, combined, or temporal). The remaining components are dropped from
consideration.
The head-motion estimates calculated in the correction step were also
placed within the corresponding confounds file.
The confound time series derived from head motion estimates and global
signals were expanded with the inclusion of temporal derivatives and
quadratic terms for each [@confounds_satterthwaite_2013].
Frames that exceeded a threshold of {regressors_fd_th} mm FD or
{regressors_dvars_th} standardized DVARS were annotated as motion outliers.
Additional nuisance timeseries are calculated by means of principal components
analysis of the signal found within a thin band (*crown*) of voxels around
the edge of the brain, as proposed by [@patriat_improved_2017].
"""
    inputnode = pe.Node(
        niu.IdentityInterface(
            fields=[
                'bold',
                'bold_mask',
                'movpar_file',
                'rmsd_file',
                'skip_vols',
                'anat_mask',
                'anat_tpms',
                'boldref2anat_xfm',
            ]
        ),
        name='inputnode',
    )
    outputnode = pe.Node(
        niu.IdentityInterface(
            fields=[
                'confounds_file',
                'confounds_metadata',
                'acompcor_masks',
                'tcompcor_mask',
                'crown_mask',
            ]
        ),
        name='outputnode',
    )

    # Project anat mask into BOLD space and merge with BOLD brainmask
    anat_mask_tfm = pe.Node(
        ApplyTransforms(interpolation='MultiLabel', invert_transform_flags=[True]),
        name='anat_mask_tfm',
    )
    union_mask = pe.Node(niu.Function(function=_binary_union), name='union_mask')

    # Create the crown mask
    dilated_mask = pe.Node(BinaryDilation(), name='dilated_mask')
    subtract_mask = pe.Node(BinarySubtraction(), name='subtract_mask')

    # DVARS
    dvars = pe.Node(
        nac.ComputeDVARS(save_nstd=True, save_std=True, remove_zerovariance=True),
        name='dvars',
        mem_gb=mem_gb,
    )

    # Frame displacement
    fdisp = pe.Node(nac.FramewiseDisplacement(parameter_source='SPM'), name='fdisp', mem_gb=mem_gb)

    # Generate aCompCor probseg maps
    acc_masks = pe.Node(aCompCorMasks(is_aseg=freesurfer), name='acc_masks')

    # Resample probseg maps in BOLD space via BOLD-to-anat transform
    acc_msk_tfm = pe.MapNode(
        ApplyTransforms(interpolation='Gaussian', invert_transform_flags=[True]),
        iterfield=['input_image'],
        name='acc_msk_tfm',
        mem_gb=0.1,
    )
    acc_msk_brain = pe.MapNode(ApplyMask(), name='acc_msk_brain', iterfield=['in_file'])
    acc_msk_bin = pe.MapNode(Binarize(thresh_low=0.99), name='acc_msk_bin', iterfield=['in_file'])
    acompcor = pe.Node(
        ACompCor(
            components_file='acompcor.tsv',
            header_prefix='a_comp_cor_',
            pre_filter='cosine',
            save_pre_filter=True,
            save_metadata=True,
            mask_names=['CSF', 'WM', 'combined'],
            merge_method='none',
            failure_mode='NaN',
        ),
        name='acompcor',
        mem_gb=mem_gb,
    )

    crowncompcor = pe.Node(
        ACompCor(
            components_file='crown_compcor.tsv',
            header_prefix='edge_comp_',
            pre_filter='cosine',
            save_pre_filter=True,
            save_metadata=True,
            mask_names=['Edge'],
            merge_method='none',
            failure_mode='NaN',
            num_components=24,
        ),
        name='crowncompcor',
        mem_gb=mem_gb,
    )

    tcompcor = pe.Node(
        TCompCor(
            components_file='tcompcor.tsv',
            header_prefix='t_comp_cor_',
            pre_filter='cosine',
            save_pre_filter=True,
            save_metadata=True,
            percentile_threshold=0.02,
            failure_mode='NaN',
        ),
        name='tcompcor',
        mem_gb=mem_gb,
    )

    # Set number of components
    if regressors_all_comps:
        acompcor.inputs.num_components = 'all'
        tcompcor.inputs.num_components = 'all'
    else:
        acompcor.inputs.variance_threshold = 0.5
        tcompcor.inputs.variance_threshold = 0.5

    # Set TR if present
    if 'RepetitionTime' in metadata:
        tcompcor.inputs.repetition_time = metadata['RepetitionTime']
        acompcor.inputs.repetition_time = metadata['RepetitionTime']
        crowncompcor.inputs.repetition_time = metadata['RepetitionTime']

    # Split aCompCor results into a_comp_cor, c_comp_cor, w_comp_cor
    rename_acompcor = pe.Node(RenameACompCor(), name='rename_acompcor')

    # Global and segment regressors
    signals_class_labels = [
        'global_signal',
        'csf',
        'white_matter',
        'csf_wm',
        'tcompcor',
    ]
    merge_rois = pe.Node(
        niu.Merge(3, ravel_inputs=True), name='merge_rois', run_without_submitting=True
    )
    signals = pe.Node(
        SignalExtraction(class_labels=signals_class_labels), name='signals', mem_gb=mem_gb
    )

    # Arrange confounds
    add_dvars_header = pe.Node(
        AddTSVHeader(columns=['dvars']),
        name='add_dvars_header',
        mem_gb=0.01,
        run_without_submitting=True,
    )
    add_std_dvars_header = pe.Node(
        AddTSVHeader(columns=['std_dvars']),
        name='add_std_dvars_header',
        mem_gb=0.01,
        run_without_submitting=True,
    )
    add_motion_headers = pe.Node(
        AddTSVHeader(columns=['trans_x', 'trans_y', 'trans_z', 'rot_x', 'rot_y', 'rot_z']),
        name='add_motion_headers',
        mem_gb=0.01,
        run_without_submitting=True,
    )
    add_rmsd_header = pe.Node(
        AddTSVHeader(columns=['rmsd']),
        name='add_rmsd_header',
        mem_gb=0.01,
        run_without_submitting=True,
    )
    concat = pe.Node(GatherConfounds(), name='concat', mem_gb=0.01, run_without_submitting=True)

    # CompCor metadata
    tcc_metadata_filter = pe.Node(FilterDropped(), name='tcc_metadata_filter')
    acc_metadata_filter = pe.Node(FilterDropped(), name='acc_metadata_filter')
    tcc_metadata_fmt = pe.Node(
        TSV2JSON(
            index_column='component',
            drop_columns=['mask'],
            output=None,
            additional_metadata={'Method': 'tCompCor'},
            enforce_case=True,
        ),
        name='tcc_metadata_fmt',
    )
    acc_metadata_fmt = pe.Node(
        TSV2JSON(
            index_column='component',
            output=None,
            additional_metadata={'Method': 'aCompCor'},
            enforce_case=True,
        ),
        name='acc_metadata_fmt',
    )
    crowncc_metadata_fmt = pe.Node(
        TSV2JSON(
            index_column='component',
            output=None,
            additional_metadata={'Method': 'EdgeRegressor'},
            enforce_case=True,
        ),
        name='crowncc_metadata_fmt',
    )

    # Combine all confounds metadata
    mrg_conf_metadata = pe.Node(
        niu.Merge(4), name='merge_confound_metadata', run_without_submitting=True
    )
    # Tissue mean time series
    mrg_conf_metadata.inputs.in3 = {label: {'Method': 'Mean'} for label in signals_class_labels}
    # Movement parameters
    mrg_conf_metadata.inputs.in4 = {
        'trans_x': {'Description': 'Translation along left-right axis.', 'Units': 'mm'},
        'trans_y': {'Description': 'Translation along anterior-posterior axis.', 'Units': 'mm'},
        'trans_z': {'Description': 'Translation along superior-inferior axis.', 'Units': 'mm'},
        'rot_x': {
            'Description': 'Rotation about left-right axis. Also known as "pitch".',
            'Units': 'rad',
        },
        'rot_y': {
            'Description': 'Rotation about anterior-posterior axis. Also known as "roll".',
            'Units': 'rad',
        },
        'rot_z': {
            'Description': 'Rotation about superior-inferior axis. Also known as "yaw".',
            'Units': 'rad',
        },
        'framewise_displacement': {'Units': 'mm'},
    }
    mrg_conf_metadata2 = pe.Node(
        DictMerge(), name='merge_confound_metadata2', run_without_submitting=True
    )

    # Expand model to include derivatives and quadratics
    model_expand = pe.Node(
        ExpandModel(model_formula='(dd1(rps + wm + csf + gsr))^^2 + others'),
        name='model_expansion',
    )

    # Add spike regressors
    spike_regress = pe.Node(
        SpikeRegressors(fd_thresh=regressors_fd_th, dvars_thresh=regressors_dvars_th),
        name='spike_regressors',
    )

    # Generate reportlet (ROIs)
    mrg_compcor = pe.Node(
        niu.Merge(3, ravel_inputs=True), name='mrg_compcor', run_without_submitting=True
    )
    rois_plot = pe.Node(
        ROIsPlot(colors=['b', 'magenta', 'g'], generate_report=True),
        name='rois_plot',
        mem_gb=mem_gb,
    )

    ds_report_bold_rois = pe.Node(
        DerivativesDataSink(
            desc='rois',
            datatype='figures',
            dismiss_entities=DEFAULT_DISMISS_ENTITIES,
        ),
        name='ds_report_bold_rois',
        run_without_submitting=True,
        mem_gb=DEFAULT_MEMORY_MIN_GB,
    )

    # Generate reportlet (CompCor)
    mrg_cc_metadata = pe.Node(
        niu.Merge(2), name='merge_compcor_metadata', run_without_submitting=True
    )
    compcor_plot = pe.Node(
        CompCorVariancePlot(
            variance_thresholds=(0.5, 0.7, 0.9),
            metadata_sources=['tCompCor', 'aCompCor', 'crownCompCor'],
        ),
        name='compcor_plot',
    )

    ds_report_compcor = pe.Node(
        DerivativesDataSink(
            desc='compcorvar',
            datatype='figures',
            dismiss_entities=DEFAULT_DISMISS_ENTITIES,
        ),
        name='ds_report_compcor',
        run_without_submitting=True,
        mem_gb=DEFAULT_MEMORY_MIN_GB,
    )

    # Generate reportlet (Confound correlation)
    conf_corr_plot = pe.Node(
        ConfoundsCorrelationPlot(reference_column='global_signal', max_dim=20),
        name='conf_corr_plot',
    )
    ds_report_conf_corr = pe.Node(
        DerivativesDataSink(
            desc='confoundcorr',
            datatype='figures',
            dismiss_entities=DEFAULT_DISMISS_ENTITIES,
        ),
        name='ds_report_conf_corr',
        run_without_submitting=True,
        mem_gb=DEFAULT_MEMORY_MIN_GB,
    )

    def _last(inlist):
        return inlist[-1]

    def _select_cols(table):
        import pandas as pd

        return [
            col
            for col in pd.read_table(table, nrows=2).columns
            if not col.startswith(('a_comp_cor_', 't_comp_cor_', 'std_dvars'))
        ]

    # fmt:off
    workflow.connect([
        # connect inputnode to each non-anatomical confound node
        (inputnode, dvars, [('bold', 'in_file'),
                            ('bold_mask', 'in_mask')]),
        (inputnode, fdisp, [('movpar_file', 'in_file')]),
        # Brain mask
        (inputnode, anat_mask_tfm, [('anat_mask', 'input_image'),
                                   ('bold_mask', 'reference_image'),
                                   ('boldref2anat_xfm', 'transforms')]),
        (inputnode, union_mask, [('bold_mask', 'mask1')]),
        (anat_mask_tfm, union_mask, [('output_image', 'mask2')]),
        (union_mask, dilated_mask, [('out', 'in_mask')]),
        (union_mask, subtract_mask, [('out', 'in_subtract')]),
        (dilated_mask, subtract_mask, [('out_mask', 'in_base')]),
        (subtract_mask, outputnode, [('out_mask', 'crown_mask')]),
        # aCompCor
        (inputnode, acompcor, [('bold', 'realigned_file'),
                               ('skip_vols', 'ignore_initial_volumes')]),
        (inputnode, acc_masks, [('anat_tpms', 'in_vfs'),
                                (('bold', _get_zooms), 'bold_zooms')]),
        (inputnode, acc_msk_tfm, [('boldref2anat_xfm', 'transforms'),
                                  ('bold_mask', 'reference_image')]),
        (inputnode, acc_msk_brain, [('bold_mask', 'in_mask')]),
        (acc_masks, acc_msk_tfm, [('out_masks', 'input_image')]),
        (acc_msk_tfm, acc_msk_brain, [('output_image', 'in_file')]),
        (acc_msk_brain, acc_msk_bin, [('out_file', 'in_file')]),
        (acc_msk_bin, acompcor, [('out_file', 'mask_files')]),
        (acompcor, rename_acompcor, [('components_file', 'components_file'),
                                     ('metadata_file', 'metadata_file')]),

        # crownCompCor
        (inputnode, crowncompcor, [('bold', 'realigned_file'),
                                   ('skip_vols', 'ignore_initial_volumes')]),
        (subtract_mask, crowncompcor, [('out_mask', 'mask_files')]),

        # tCompCor
        (inputnode, tcompcor, [('bold', 'realigned_file'),
                               ('skip_vols', 'ignore_initial_volumes'),
                               ('bold_mask', 'mask_files')]),
        # Global signals extraction (constrained by anatomy)
        (inputnode, signals, [('bold', 'in_file')]),
        (inputnode, merge_rois, [('bold_mask', 'in1')]),
        (acc_msk_bin, merge_rois, [('out_file', 'in2')]),
        (tcompcor, merge_rois, [('high_variance_masks', 'in3')]),
        (merge_rois, signals, [('out', 'label_files')]),

        # Collate computed confounds together
        (inputnode, add_motion_headers, [('movpar_file', 'in_file')]),
        (inputnode, add_rmsd_header, [('rmsd_file', 'in_file')]),
        (dvars, add_dvars_header, [('out_nstd', 'in_file')]),
        (dvars, add_std_dvars_header, [('out_std', 'in_file')]),
        (signals, concat, [('out_file', 'signals')]),
        (fdisp, concat, [('out_file', 'fd')]),
        (tcompcor, concat, [('components_file', 'tcompcor'),
                            ('pre_filter_file', 'cos_basis')]),
        (rename_acompcor, concat, [('components_file', 'acompcor')]),
        (crowncompcor, concat, [('components_file', 'crowncompcor')]),
        (add_motion_headers, concat, [('out_file', 'motion')]),
        (add_rmsd_header, concat, [('out_file', 'rmsd')]),
        (add_dvars_header, concat, [('out_file', 'dvars')]),
        (add_std_dvars_header, concat, [('out_file', 'std_dvars')]),

        # Confounds metadata
        (tcompcor, tcc_metadata_filter, [('metadata_file', 'in_file')]),
        (tcc_metadata_filter, tcc_metadata_fmt, [('out_file', 'in_file')]),
        (rename_acompcor, acc_metadata_filter, [('metadata_file', 'in_file')]),
        (acc_metadata_filter, acc_metadata_fmt, [('out_file', 'in_file')]),
        (crowncompcor, crowncc_metadata_fmt, [('metadata_file', 'in_file')]),
        (tcc_metadata_fmt, mrg_conf_metadata, [('output', 'in1')]),
        (acc_metadata_fmt, mrg_conf_metadata, [('output', 'in2')]),
        (crowncc_metadata_fmt, mrg_conf_metadata, [('output', 'in3')]),
        (mrg_conf_metadata, mrg_conf_metadata2, [('out', 'in_dicts')]),

        # Expand the model with derivatives, quadratics, and spikes
        (concat, model_expand, [('confounds_file', 'confounds_file')]),
        (model_expand, spike_regress, [('confounds_file', 'confounds_file')]),

        # Set outputs
        (spike_regress, outputnode, [('confounds_file', 'confounds_file')]),
        (mrg_conf_metadata2, outputnode, [('out_dict', 'confounds_metadata')]),
        (tcompcor, outputnode, [('high_variance_masks', 'tcompcor_mask')]),
        (acc_msk_bin, outputnode, [('out_file', 'acompcor_masks')]),
        (inputnode, rois_plot, [('bold', 'in_file'),
                                ('bold_mask', 'in_mask')]),
        (tcompcor, mrg_compcor, [('high_variance_masks', 'in1')]),
        (acc_msk_bin, mrg_compcor, [(('out_file', _last), 'in2')]),
        (subtract_mask, mrg_compcor, [('out_mask', 'in3')]),
        (mrg_compcor, rois_plot, [('out', 'in_rois')]),
        (rois_plot, ds_report_bold_rois, [('out_report', 'in_file')]),
        (tcompcor, mrg_cc_metadata, [('metadata_file', 'in1')]),
        (acompcor, mrg_cc_metadata, [('metadata_file', 'in2')]),
        (crowncompcor, mrg_cc_metadata, [('metadata_file', 'in3')]),
        (mrg_cc_metadata, compcor_plot, [('out', 'metadata_files')]),
        (compcor_plot, ds_report_compcor, [('out_file', 'in_file')]),
        (inputnode, conf_corr_plot, [('skip_vols', 'ignore_initial_volumes')]),
        (concat, conf_corr_plot, [('confounds_file', 'confounds_file'),
                                  (('confounds_file', _select_cols), 'columns')]),
        (conf_corr_plot, ds_report_conf_corr, [('out_file', 'in_file')]),
    ])
    # fmt: on

    return workflow


def init_carpetplot_wf(
    mem_gb: float, metadata: dict, cifti_output: bool, name: str = 'bold_carpet_wf'
):
    """
    Build a workflow to generate *carpet* plots.

    Resamples the MNI parcellation (ad-hoc parcellation derived from the
    Harvard-Oxford template and others).

    Parameters
    ----------
    mem_gb : :obj:`float`
        Size of BOLD file in GB - please note that this size
        should be calculated after resamplings that may extend
        the FoV
    metadata : :obj:`dict`
        BIDS metadata for BOLD file
    name : :obj:`str`
        Name of workflow (default: ``bold_carpet_wf``)

    Inputs
    ------
    bold
        BOLD image, after the prescribed corrections (STC, HMC and SDC)
        when available.
    bold_mask
        BOLD series mask
    confounds_file
        TSV of all aggregated confounds
    boldref2anat_xfm
        Affine matrix that maps the BOLD reference space into alignment with
        the anatomical space
    std2anat_xfm
        ANTs-compatible affine-and-warp transform file
    cifti_bold
        BOLD image in CIFTI format, to be used in place of volumetric BOLD
    crown_mask
        Mask of brain edge voxels
    acompcor_mask
        Mask of deep WM+CSF
    dummy_scans
        Number of nonsteady states to be dropped at the beginning of the timeseries.

    Outputs
    -------
    out_carpetplot
        Path of the generated SVG file

    """
    from niworkflows.engine.workflows import LiterateWorkflow as Workflow

    if not cifti_output:
        raise NotImplementedError(
            'Carpetplot can only be generated with a dense timeseries via `--cifti-output`.'
        )

    inputnode = pe.Node(
        niu.IdentityInterface(
            fields=[
                'bold',
                'bold_mask',
                'confounds_file',
                'boldref2anat_xfm',
                'std2anat_xfm',
                'cifti_bold',
                'crown_mask',
                'acompcor_mask',
                'dummy_scans',
            ]
        ),
        name='inputnode',
    )

    outputnode = pe.Node(niu.IdentityInterface(fields=['out_carpetplot']), name='outputnode')

    # Carpetplot and confounds plot
    conf_plot = pe.Node(
        FMRISummary(
            tr=metadata['RepetitionTime'],
            confounds_list=[
                ('global_signal', None, 'GS'),
                ('csf', None, 'CSF'),
                ('white_matter', None, 'WM'),
                ('std_dvars', None, 'DVARS'),
                ('framewise_displacement', 'mm', 'FD'),
            ],
        ),
        name='conf_plot',
        mem_gb=mem_gb,
    )
    ds_report_bold_conf = pe.Node(
        DerivativesDataSink(
            desc='carpetplot',
            datatype='figures',
            extension='svg',
            dismiss_entities=DEFAULT_DISMISS_ENTITIES,
        ),
        name='ds_report_bold_conf',
        run_without_submitting=True,
        mem_gb=DEFAULT_MEMORY_MIN_GB,
    )

    workflow = Workflow(name=name)
    workflow.connect([
        (inputnode, conf_plot, [
            ('bold', 'in_nifti'),
            ('cifti_bold', 'in_cifti'),
            ('confounds_file', 'confounds_file'),
            ('dummy_scans', 'drop_trs'),
        ]),
        (conf_plot, ds_report_bold_conf, [('out_file', 'in_file')]),
        (conf_plot, outputnode, [('out_file', 'out_carpetplot')]),
    ])  # fmt:skip
    return workflow


def _binary_union(mask1, mask2):
    """Generate the union of two masks."""
    from pathlib import Path

    import nibabel as nb
    import numpy as np

    img = nb.load(mask1)
    mskarr1 = np.asanyarray(img.dataobj, dtype=int) > 0
    mskarr2 = np.asanyarray(nb.load(mask2).dataobj, dtype=int) > 0
    out = img.__class__(mskarr1 | mskarr2, img.affine, img.header)
    out.set_data_dtype('uint8')
    out_name = Path('mask_union.nii.gz').absolute()
    out.to_filename(out_name)
    return str(out_name)


def _get_zooms(in_file):
    import nibabel as nb

    return tuple(nb.load(in_file).header.get_zooms()[:3])



import typing as ty

import nipype.interfaces.utility as niu
import nipype.pipeline.engine as pe
import templateflow.api as tf
from niworkflows.engine import Workflow, tag
from smriprep.interfaces.workbench import SurfaceResample
from smriprep.workflows.surfaces import init_morph_grayords_wf

from nibabies.config import DEFAULT_MEMORY_MIN_GB
from nibabies.data import load as load_data
from nibabies.interfaces.utils import CiftiSelect


@tag('anat.resample-surfs')
def init_anat_fsLR_resampling_wf(
    grayord_density: ty.Literal['91k'], mcribs: bool, name='anat_fsLR_resampling_wf'
) -> Workflow:
    """Resample the surfaces into fsLR space"""
    workflow = Workflow(name=name)
    fslr_density = '32k' if grayord_density == '91k' else '59k'

    workflow.__desc__ = """\
The BOLD time-series were resampled onto the left/right-symmetric template
"fsLR" [@hcppipelines].
"""

    inputnode = pe.Node(
        niu.IdentityInterface(
            fields=[
                'subject_id',
                'subjects_dir',
                'surfaces',
                'morphometrics',
                'sphere_reg',
                'sphere_reg_fsLR',
            ]
        ),
        name='inputnode',
    )

    itersource = pe.Node(
        niu.IdentityInterface(fields=['hemi']),
        name='itersource',
        iterables=[('hemi', ['L', 'R'])],
    )

    outputnode = pe.Node(
        niu.IdentityInterface(
            fields=[
                'midthickness_fsLR',
                'cifti_morph',
                'cifti_metadata',
            ]
        ),
        name='outputnode',
    )

    # select white, midthickness and pial surfaces based on hemi
    select_surfaces = pe.Node(CiftiSelect(), name='select_surfaces')

    if mcribs:
        atlases = load_data.cached('atlases')
        # use dHCP 32k fsLR instead
        select_surfaces.inputs.template_spheres = [
            str(atlases / 'tpl-dHCP_space-fsLR_hemi-L_den-32k_desc-week42_sphere.surf.gii'),
            str(atlases / 'tpl-dHCP_space-fsLR_hemi-R_den-32k_desc-week42_sphere.surf.gii'),
        ]
    else:
        select_surfaces.inputs.template_spheres = [
            str(sphere)
            for sphere in tf.get(
                template='fsLR',
                density=fslr_density,
                suffix='sphere',
                space=None,
                extension='.surf.gii',
            )
        ]

    # Line 393 of FreeSurfer2CaretConvertAndRegisterNonlinear.sh
    downsampled_midthickness = pe.Node(
        SurfaceResample(method='BARYCENTRIC'),
        name='downsampled_midthickness',
        mem_gb=DEFAULT_MEMORY_MIN_GB,
    )

    joinnode = pe.JoinNode(
        niu.IdentityInterface(fields=['midthickness_fsLR']),
        name='joinnode',
        joinsource='itersource',
    )

    # resample surfaces / morphometrics to 32k
    if mcribs:
        morph_grayords_wf = init_mcribs_morph_grayords_wf(grayord_density)

        workflow.connect([
            (inputnode, morph_grayords_wf, [
                ('morphometrics', 'inputnode.morphometrics'),
                ('surfaces', 'inputnode.surfaces'),
                ('sphere_reg_fsLR', 'inputnode.sphere_reg')]),
            (joinnode, morph_grayords_wf, [
                ('midthickness_fsLR', 'inputnode.midthickness_fsLR')]),
        ])  # fmt:skip
    else:
        morph_grayords_wf = init_morph_grayords_wf(grayord_density)

    workflow.connect([
        (inputnode, select_surfaces, [
            ('surfaces', 'surfaces'),
            ('sphere_reg_fsLR', 'spherical_registrations')]),
        (itersource, select_surfaces, [('hemi', 'hemi')]),
        # Downsample midthickness to fsLR density
        (select_surfaces, downsampled_midthickness, [
            ('midthickness', 'surface_in'),
            ('sphere_reg', 'current_sphere'),
            ('template_sphere', 'new_sphere')]),
        (downsampled_midthickness, joinnode, [('surface_out', 'midthickness_fsLR')]),
        (joinnode, outputnode, [('midthickness_fsLR', 'midthickness_fsLR')]),
        # resample morphometrics to fsLR 32k
        (inputnode, morph_grayords_wf, [
            ('subject_id', 'inputnode.subject_id'),
            ('subjects_dir', 'inputnode.subjects_dir')]),
        (morph_grayords_wf, outputnode, [
            ('outputnode.cifti_morph', 'cifti_morph'),
            ('outputnode.cifti_metadata', 'cifti_metadata')]),
    ])  # fmt:skip
    return workflow


@tag('anat.resample-morphs-grayords')
def init_mcribs_morph_grayords_wf(
    grayord_density: ty.Literal['91k'],  # Only 91k supported ATM
    name: str = 'morph_grayords_wf',
):
    """
    Sample Grayordinates files onto the fsLR atlas.

    If `mcribs` is disabled (default), the fsaverage sphere will be resampled to fsLR.
    If `mcribs` is enabled, the M-CRIB-S sphere will be resampled to dHCP 42 week.

    Outputs are in CIFTI2 format.

    Workflow Graph
        .. workflow::
            :graph2use: colored
            :simple_form: yes

            from smriprep.workflows.surfaces import init_morph_grayords_wf
            wf = init_morph_grayords_wf(grayord_density="91k")

    Parameters
    ----------
    grayord_density : :obj:`str`
        Either `91k` or `170k`, representing the total of vertices or *grayordinates*.
    name : :obj:`str`
        Unique name for the subworkflow (default: ``"morph_grayords_wf"``)

    Inputs
    ------
    subject_id : :obj:`str`
        FreeSurfer subject ID
    subjects_dir : :obj:`str`
        FreeSurfer SUBJECTS_DIR

    Outputs
    -------
    cifti_morph : :obj:`list` of :obj:`str`
        Paths of CIFTI dscalar files
    cifti_metadata : :obj:`list` of :obj:`str`
        Paths to JSON files containing metadata corresponding to ``cifti_morph``

    """
    from nipype.interfaces.workbench import MetricResample
    from smriprep.interfaces.cifti import GenerateDScalar

    from nibabies.interfaces.workbench import SurfaceVertexAreas

    workflow = Workflow(name=name)
    workflow.__desc__ = f"""\
*Grayordinate* "dscalar" files [@hcppipelines] containing {grayord_density} samples were
also generated using `M-CRIB-S` as an intermediate standardized
surface space.
"""

    inputnode = pe.Node(
        niu.IdentityInterface(
            fields=[
                'subject_id',
                'subjects_dir',
                'surfaces',
                'morphometrics',
                'sphere_reg',
                'midthickness_fsLR',
            ]
        ),
        name='inputnode',
    )

    outputnode = pe.Node(
        niu.IdentityInterface(fields=['cifti_morph', 'cifti_metadata']),
        name='outputnode',
    )

    surfmorph_list = pe.Node(
        niu.Merge(3, ravel_inputs=True),
        name='surfmorph_list',
        run_without_submitting=True,
    )

    subject_midthickness = pe.Node(
        niu.Function(function=_get_surf),
        name='get_midthickness',
        run_without_submitting=True,
    )
    subject_midthickness.inputs.name = 'midthickness'
    subject_midthickness.inputs.mult = 3

    template_midthickness = subject_midthickness.clone('get_new_midthickness')

    # Create Vertex Areas from midthickness surfaces
    subject_va = pe.MapNode(SurfaceVertexAreas(), iterfield='in_file', name='subject_va')
    template_va = subject_va.clone('template_va')

    # Setup Workbench command. LR ordering for hemi can be assumed, as it is imposed
    # by the iterfield of the MapNode in the surface sampling workflow above.
    resample = pe.MapNode(
        MetricResample(method='ADAP_BARY_AREA', area_metrics=True),
        name='resample',
        iterfield=[
            'in_file',
            'out_file',
            'new_sphere',
            'new_area',
            'current_sphere',
            'current_area',
        ],
    )

    atlases = load_data.cached('atlases')
    resample.inputs.new_sphere = [  # 32k
        str(atlases / 'tpl-dHCP_space-fsLR_hemi-L_den-32k_desc-week42_sphere.surf.gii'),
        str(atlases / 'tpl-dHCP_space-fsLR_hemi-R_den-32k_desc-week42_sphere.surf.gii'),
    ] * 3
    resample.inputs.out_file = [
        f'space-fsLR_hemi-{h}_den-{grayord_density}_{morph}.shape.gii'
        # Order: curv-L, curv-R, sulc-L, sulc-R, thickness-L, thickness-R
        for morph in ('curv', 'sulc', 'thickness')
        for h in 'LR'
    ]

    gen_cifti = pe.MapNode(
        GenerateDScalar(
            grayordinates=grayord_density,
        ),
        iterfield=['scalar_name', 'scalar_surfs'],
        name='gen_cifti',
    )
    gen_cifti.inputs.scalar_name = ['curv', 'sulc', 'thickness']

    # fmt: off
    workflow.connect([
        (inputnode, resample, [(('sphere_reg', _triple), 'current_sphere')]),
        (inputnode, subject_midthickness, [('surfaces', 'surfaces')]),
        (inputnode, template_midthickness, [('midthickness_fsLR', 'surfaces')]),
        (subject_midthickness, subject_va, [('out', 'in_file')]),
        (template_midthickness, template_va, [('out', 'in_file')]),
        (subject_va, resample, [('out_file', 'current_area')]),
        (template_va, resample, [('out_file', 'new_area')]),
        (inputnode, surfmorph_list, [
            (('morphometrics', _get_surf, 'curv'), 'in1'),
            (('morphometrics', _get_surf, 'sulc'), 'in2'),
            (('morphometrics', _get_surf, 'thickness'), 'in3'),
        ]),
        (surfmorph_list, resample, [('out', 'in_file')]),
        (resample, gen_cifti, [
            (('out_file', _collate), 'scalar_surfs')]),
        (gen_cifti, outputnode, [('out_file', 'cifti_morph'),
                                 ('out_metadata', 'cifti_metadata')]),
    ])
    # fmt: on
    return workflow


def _get_surf(surfaces, name, mult=1):
    from smriprep.workflows.surfaces import _sorted_by_basename

    'Select a specific surface by name, and optionally multiple it.'
    if not surfaces:
        return surfaces
    return [surf for surf in _sorted_by_basename(surfaces) if name in surf] * mult


def _triple(in_list):
    return in_list * 3


def _collate(files):
    return [files[i : i + 2] for i in range(0, len(files), 2)]



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
#
# Copyright The NiPreps Developers <nipreps@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# We support and encourage derived works from this project, please read
# about our expectations at
#
#     https://www.nipreps.org/community/licensing/
#
# STATEMENT OF CHANGES: This file is derived from sources licensed under the Apache-2.0 terms,
# and uses the following portion of the original code:
# https://github.com/dandi/dandi-cli/blob/da3b7a726c4a352dfb53a0c6bee59e660de827e6/dandi/utils.py#L49-L82
#
#
# ORIGINAL WORK'S ATTRIBUTION NOTICE:
#
#    Copyright DANDI Client Developers
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
import sys


def is_interactive():
    """Return True if all in/outs are tty"""
    # TODO: check on windows if hasattr check would work correctly and add value:
    #
    return sys.stdin.isatty() and sys.stdout.isatty() and sys.stderr.isatty()


def setup_exceptionhook(ipython=False):
    """Overloads default sys.excepthook with our exceptionhook handler.
    If interactive, our exceptionhook handler will invoke
    pdb.post_mortem; if not interactive, then invokes default handler.
    """

    def _pdb_excepthook(type, value, tb):
        import traceback

        traceback.print_exception(type, value, tb)
        print()
        if is_interactive():
            import pdb

            pdb.post_mortem(tb)

    if ipython:
        from IPython.core import ultratb

        sys.excepthook = ultratb.FormattedTB(
            mode='Verbose',
            # color_scheme='Linux',
            call_pdb=is_interactive(),
        )
    else:
        sys.excepthook = _pdb_excepthook



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
#
# Copyright The NiPreps Developers <nipreps@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# We support and encourage derived works from this project, please read
# about our expectations at
#
#     https://www.nipreps.org/community/licensing/
#
"""Test parser."""

from argparse import ArgumentError
from contextlib import nullcontext

import pytest
from packaging.version import Version

from nibabies import config
from nibabies.cli import version as _version
from nibabies.cli.parser import _build_parser, parse_args
from nibabies.tests.test_config import _reset_config

MIN_ARGS = ['data/', 'out/', 'participant']


@pytest.mark.parametrize(
    ('args', 'code'),
    [
        ([], 2),
        (MIN_ARGS, 2),  # bids_dir does not exist
        (MIN_ARGS + ['--fs-license-file'], 2),
        (MIN_ARGS + ['--fs-license-file', 'fslicense.txt'], 2),
    ],
)
def test_parser_errors(args, code):
    """Check behavior of the parser."""
    with pytest.raises(SystemExit) as error:
        _build_parser().parse_args(args)

    assert error.value.code == code


@pytest.mark.parametrize('args', [MIN_ARGS, MIN_ARGS + ['--fs-license-file']])
def test_parser_valid(tmp_path, args):
    """Check valid arguments."""
    datapath = tmp_path / 'data'
    datapath.mkdir(exist_ok=True)
    args[0] = str(datapath)

    if '--fs-license-file' in args:
        _fs_file = tmp_path / 'license.txt'
        _fs_file.write_text('')
        args.insert(args.index('--fs-license-file') + 1, str(_fs_file.absolute()))

    opts = _build_parser().parse_args(args)

    assert opts.bids_dir == datapath


@pytest.mark.parametrize(
    ('argval', 'gb'),
    [
        ('1G', 1),
        ('1GB', 1),
        ('1000', 1),  # Default units are MB
        ('32000', 32),  # Default units are MB
        ('4000', 4),  # Default units are MB
        ('1000M', 1),
        ('1000MB', 1),
        ('1T', 1000),
        ('1TB', 1000),
        (f'{1e6:.0f}K', 1),
        (f'{1e6:.0f}KB', 1),
        (f'{1e9:.0f}B', 1),
    ],
)
def test_memory_arg(tmp_path, argval, gb):
    """Check the correct parsing of the memory argument."""
    datapath = tmp_path / 'data'
    datapath.mkdir(exist_ok=True)
    _fs_file = tmp_path / 'license.txt'
    _fs_file.write_text('')

    args = [str(datapath)] + MIN_ARGS[1:] + ['--fs-license-file', str(_fs_file), '--mem', argval]
    opts = _build_parser().parse_args(args)

    assert opts.memory_gb == gb


@pytest.mark.parametrize(('current', 'latest'), [('1.0.0', '1.3.2'), ('1.3.2', '1.3.2')])
def test_get_parser_update(monkeypatch, capsys, current, latest):
    """Make sure the out-of-date banner is shown."""
    expectation = Version(current) < Version(latest)

    def _mock_check_latest(*args, **kwargs):
        return Version(latest)

    monkeypatch.setattr(config.environment, 'version', current)
    monkeypatch.setattr(_version, 'check_latest', _mock_check_latest)

    _build_parser()
    captured = capsys.readouterr().err

    msg = f"""\
You are using NiBabies-{current}, and a newer version of NiBabies is available: {latest}.
Please check out our documentation about how and when to upgrade:
https://fmriprep.readthedocs.io/en/latest/faq.html#upgrading"""

    assert (msg in captured) is expectation


@pytest.mark.parametrize('flagged', [(True, None), (True, 'random reason'), (False, None)])
def test_get_parser_blacklist(monkeypatch, capsys, flagged):
    """Make sure the blacklisting banner is shown."""

    def _mock_is_bl(*args, **kwargs):
        return flagged

    monkeypatch.setattr(_version, 'is_flagged', _mock_is_bl)

    _build_parser()
    captured = capsys.readouterr().err

    assert ('FLAGGED' in captured) is flagged[0]
    if flagged[0]:
        assert (flagged[1] or 'reason: unknown') in captured


def test_parse_args(tmp_path, minimal_bids):
    """Basic smoke test showing that our parse_args() function
    implements the BIDS App protocol"""
    out_dir = tmp_path / 'out'
    work_dir = tmp_path / 'work'

    parse_args(
        args=[
            str(minimal_bids),
            str(out_dir),
            'participant',  # BIDS App
            '-w',
            str(work_dir),  # Don't pollute CWD
            '--skip-bids-validation',  # Empty files make BIDS sad
        ]
    )
    assert config.execution.layout.root == str(minimal_bids)
    _reset_config()


def test_bids_filter_file(tmp_path, capsys):
    bids_path = tmp_path / 'data'
    out_path = tmp_path / 'out'
    bff = tmp_path / 'filter.json'
    args = [str(bids_path), str(out_path), 'participant', '--bids-filter-file', str(bff)]
    bids_path.mkdir()

    parser = _build_parser()

    with pytest.raises(SystemExit):
        parser.parse_args(args)

    err = capsys.readouterr().err
    assert 'Path does not exist:' in err

    bff.write_text('{"invalid json": }')

    with pytest.raises(SystemExit):
        parser.parse_args(args)

    err = capsys.readouterr().err
    assert 'JSON syntax error in:' in err
    _reset_config()


@pytest.mark.parametrize('st_ref', (None, '0', '1', '0.5', 'start', 'middle'))  # noqa: PT007
def test_slice_time_ref(tmp_path, st_ref):
    bids_path = tmp_path / 'data'
    out_path = tmp_path / 'out'
    args = [str(bids_path), str(out_path), 'participant']
    if st_ref:
        args.extend(['--slice-time-ref', st_ref])
    bids_path.mkdir()

    parser = _build_parser()

    parser.parse_args(args)
    _reset_config()


@pytest.mark.parametrize(
    ('args', 'expectation'),
    [
        ([], False),
        (['--use-syn-sdc'], 'error'),
        (['--use-syn-sdc', 'error'], 'error'),
        (['--use-syn-sdc', 'warn'], 'warn'),
        (['--use-syn-sdc', 'other'], (SystemExit, ArgumentError)),
    ],
)
def test_use_syn_sdc(tmp_path, args, expectation):
    bids_path = tmp_path / 'data'
    out_path = tmp_path / 'out'
    args = [str(bids_path), str(out_path), 'participant'] + args
    bids_path.mkdir()

    parser = _build_parser()

    cm = nullcontext()
    if isinstance(expectation, tuple):
        cm = pytest.raises(expectation)

    with cm:
        opts = parser.parse_args(args)

    if not isinstance(expectation, tuple):
        assert opts.use_syn_sdc == expectation

    _reset_config()


def test_derivatives(tmp_path):
    """Check the correct parsing of the derivatives argument."""
    bids_path = tmp_path / 'data'
    out_path = tmp_path / 'out'
    args = [str(bids_path), str(out_path), 'participant']
    bids_path.mkdir()

    parser = _build_parser()

    # Providing --derivatives without a path should raise an error
    temp_args = args + ['--derivatives']
    with pytest.raises((SystemExit, ArgumentError)):
        parser.parse_args(temp_args)
    _reset_config()

    # Providing --derivatives without names should automatically label them
    temp_args = args + ['--derivatives', str(bids_path / 'derivatives/smriprep')]
    opts = parser.parse_args(temp_args)
    assert opts.derivatives == {'smriprep': bids_path / 'derivatives/smriprep'}
    _reset_config()

    # Providing --derivatives with names should use them
    temp_args = args + [
        '--derivatives',
        f'anat={str(bids_path / "derivatives/smriprep")}',
    ]
    opts = parser.parse_args(temp_args)
    assert opts.derivatives == {'anat': bids_path / 'derivatives/smriprep'}
    _reset_config()

    # Providing multiple unlabeled derivatives with the same name should raise an error
    temp_args = args + [
        '--derivatives',
        str(bids_path / 'derivatives_01/smriprep'),
        str(bids_path / 'derivatives_02/smriprep'),
    ]
    with pytest.raises(ValueError, match='Received duplicate derivative name'):
        parser.parse_args(temp_args)

    _reset_config()


@pytest.mark.parametrize(
    ('args', 'expectation'),
    [
        ([], 'auto'),
        (['--surface-recon-method', 'auto'], 'auto'),
        (['--surface-recon-method', 'mcribs'], 'mcribs'),
        (['--surface-recon-method', 'infantfs'], 'infantfs'),
        (['--surface-recon-method', 'freesurfer'], 'freesurfer'),
        (['--surface-recon-method', 'none'], None),
        (['--surface-recon-method', 'None'], None),
        (['--surface-recon-method', 123], (TypeError,)),
    ],
)
def test_surface_recon_method(tmp_path, args, expectation):
    """Check the correct parsing of the memory argument."""
    datapath = tmp_path / 'data'
    datapath.mkdir(exist_ok=True)
    _fs_file = tmp_path / 'license.txt'
    _fs_file.write_text('')

    args = [str(datapath)] + MIN_ARGS[1:] + ['--fs-license-file', str(_fs_file)] + args

    cm = nullcontext()
    if isinstance(expectation, tuple):
        cm = pytest.raises(expectation)

    with cm:
        opts = _build_parser().parse_args(args)
        assert opts.surface_recon_method == expectation



from nibabel.optpkg import optional_package

from .. import __version__, config

migas = optional_package('migas')[0]


def setup_migas(init_ping: bool = True, exit_ping: bool = True) -> None:
    """
    Prepare the migas python client to communicate with a migas server.
    If ``init`` is ``True``, send an initial breadcrumb.
    """
    # generate session UUID from generated run UUID
    session_id = None
    if config.execution.run_uuid:
        session_id = config.execution.run_uuid.split('_', 1)[-1]

    migas.setup(session_id=session_id)
    if init_ping:
        # send initial status ping
        send_crumb(status='R', status_desc='workflow start')
    if exit_ping:
        from migas.error.nipype import node_execution_error

        migas.track_exit(
            'nipreps/nibabies',
            __version__,
            {'NodeExecutionError': node_execution_error},
        )


def send_crumb(**kwargs) -> dict:
    """
    Communicate with the migas telemetry server. This requires `migas.setup()` to be called.
    """
    return migas.add_breadcrumb('nipreps/nibabies', __version__, **kwargs)



import nipype.interfaces.utility as niu
import nipype.pipeline.engine as pe


def init_infant_epi_reference_wf(
    omp_nthreads: int,
    is_sbref: bool = False,
    start_frame: int = 17,
    name: str = 'infant_epi_reference_wf',
) -> pe.Workflow:
    """
    Workflow to generate a reference map from one or more infant EPI images.

    If any single-band references are provided, the reference map will be calculated from those.

    If no single-band references are provided, the BOLD files are used.
    To account for potential increased motion on the start of image acquisition, this
    workflow discards a bigger chunk of the initial frames.

    Parameters
    ----------
    omp_nthreads
        Maximum number of threads an individual process may use
    is_sbref
        A single-band reference is provided.
    start_frame
        BOLD frame to start creating the reference map from. Any earlier frames are discarded.

    Inputs
    ------
    bold_file
        BOLD EPI file
    sbref_file
        single-band reference EPI

    Outputs
    -------
    boldref_file
        The generated reference map
    boldref_mask
        Binary brain mask of the ``boldref_file``
    boldref_xfm
        Rigid-body transforms in LTA format

    """
    from niworkflows.workflows.epi.refmap import init_epi_reference_wf
    from sdcflows.interfaces.brainmask import BrainExtraction

    wf = pe.Workflow(name=name)

    inputnode = pe.Node(
        niu.IdentityInterface(fields=['epi_file']),
        name='inputnode',
    )
    outputnode = pe.Node(
        niu.IdentityInterface(fields=['boldref_file', 'boldref_mask']),
        name='outputnode',
    )

    epi_reference_wf = init_epi_reference_wf(omp_nthreads, auto_bold_nss=False)

    boldref_mask = pe.Node(BrainExtraction(), name='boldref_mask')

    # fmt:off
    wf.connect([
        (inputnode, epi_reference_wf, [('epi_file', 'inputnode.in_files')]),
        (epi_reference_wf, boldref_mask, [('outputnode.epi_ref_file', 'in_file')]),
        (epi_reference_wf, outputnode, [('outputnode.epi_ref_file', 'boldref_file')]),
        (boldref_mask, outputnode, [('out_mask', 'boldref_mask')]),
    ])
    # fmt:on
    if not is_sbref:
        select_frames = pe.Node(
            niu.Function(function=_select_frames, output_names=['t_masks']),
            name='select_frames',
        )
        select_frames.inputs.start_frame = start_frame
        # fmt:off
        wf.connect([
            (inputnode, select_frames, [('epi_file', 'in_file')]),
            (select_frames, epi_reference_wf, [('t_masks', 'inputnode.t_masks')]),
        ])
        # fmt:on
    else:
        # Won't be used but needed to placate iternode
        # To consider: Add a check to ensure this is a 3D file
        epi_reference_wf.inputs.inputnode.t_masks = [True]
    return wf


def _select_frames(in_file: str, start_frame: int) -> list:
    import nibabel as nb
    import numpy as np

    img = nb.load(in_file)
    img_len = img.shape[3]
    if start_frame >= img_len:
        start_frame = img_len - 1
    t_mask = np.array([False] * img_len, dtype=bool)
    t_mask[start_frame:] = True
    return list(t_mask)



from pathlib import Path
from shutil import copytree
from tempfile import TemporaryDirectory

import pytest

try:
    from contextlib import chdir as _chdir
except ImportError:  # PY310
    import os
    from contextlib import contextmanager

    @contextmanager  # type: ignore
    def _chdir(path):
        cwd = os.getcwd()
        os.chdir(path)
        try:
            yield
        finally:
            os.chdir(cwd)


DATA_FILES = (
    'functional.nii',
    'anatomical.nii',
    'func.dlabel.nii',
    'func.dtseries.nii',
    'epi.nii',
    'T1w.nii',
    'func_to_struct.mat',
    'atlas.nii',
    'label_list.txt',
    'sub-01_run-01_echo-1_bold.nii.gz',
    'sub-01_run-01_echo-2_bold.nii.gz',
    'sub-01_run-01_echo-3_bold.nii.gz',
    'xfm0.h5',
    'xfm1.h5',
)


@pytest.fixture(scope='package')
def data_dir():
    with TemporaryDirectory() as tmpdir:
        tmp_path = Path(tmpdir)
        for fname in DATA_FILES:
            Path.touch(tmp_path / fname)
        yield tmp_path


@pytest.fixture(autouse=True)
def _docdir(data_dir, request, tmp_path):
    # Trigger ONLY for the doctests.
    doctest_plugin = request.config.pluginmanager.getplugin('doctest')
    if isinstance(request.node, doctest_plugin.DoctestItem):
        copytree(data_dir, tmp_path, dirs_exist_ok=True)

        # Chdir only for the duration of the test.
        with _chdir(tmp_path):
            yield

    else:
        # For normal tests, we have to yield, since this is a yield-fixture.
        yield



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
#
# Copyright The NiPreps Developers <nipreps@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# We support and encourage derived works from this project, please read
# about our expectations at
#
#     https://www.nipreps.org/community/licensing/
#
from nipype.interfaces import utility as niu
from nipype.pipeline import engine as pe
from niworkflows.engine.workflows import LiterateWorkflow as Workflow
from niworkflows.interfaces.header import ValidateImage
from niworkflows.utils.misc import pass_dummy_scans

DEFAULT_MEMORY_MIN_GB = 0.01


def init_raw_boldref_wf(
    bold_file: str | None = None,
    multiecho: bool = False,
    ref_frame_start: int = 16,
    name: str = 'raw_boldref_wf',
):
    """
    Build a workflow that generates reference BOLD images for a series.

    The raw reference image is the target of :abbr:`HMC (head motion correction)`, and a
    contrast-enhanced reference is the subject of distortion correction, as well as
    boundary-based registration to anat and template spaces.

    This workflow assumes only one BOLD file has been passed.

    Workflow Graph
        .. workflow::
            :graph2use: orig
            :simple_form: yes

            from fmriprep.workflows.bold.reference import init_raw_boldref_wf
            wf = init_raw_boldref_wf()

    Parameters
    ----------
    bold_file : :obj:`str`
        BOLD series NIfTI file
    multiecho : :obj:`bool`
        If multiecho data was supplied, data from the first echo will be selected
    ref_frame_start: :obj:`int`
        BOLD frame to start creating the reference map from.
    name : :obj:`str`
        Name of workflow (default: ``raw_boldref_wf``)

    Inputs
    ------
    bold_file : str
        BOLD series NIfTI file
    dummy_scans : int or None
        Number of non-steady-state volumes specified by user at beginning of ``bold_file``

    Outputs
    -------
    bold_file : str
        Validated BOLD series NIfTI file
    boldref : str
        Reference image to which BOLD series is motion corrected
    skip_vols : int
        Number of non-steady-state volumes selected at beginning of ``bold_file``
    algo_dummy_scans : int
        Number of non-steady-state volumes agorithmically detected at
        beginning of ``bold_file``

    """
    from niworkflows.interfaces.bold import NonsteadyStatesDetector
    from niworkflows.interfaces.images import RobustAverage

    workflow = Workflow(name=name)
    workflow.__desc__ = f"""\
First, a reference volume was generated{' from the shortest echo of the BOLD run' * multiecho},
using a custom methodology of *NiBabies*, for use in head motion correction.
"""

    inputnode = pe.Node(
        niu.IdentityInterface(fields=['bold_file', 'dummy_scans']),
        name='inputnode',
    )
    outputnode = pe.Node(
        niu.IdentityInterface(
            fields=[
                'bold_file',
                'boldref',
                'skip_vols',
                'algo_dummy_scans',
                'validation_report',
            ]
        ),
        name='outputnode',
    )

    # Simplify manually setting input image
    if bold_file is not None:
        inputnode.inputs.bold_file = bold_file

    val_bold = pe.Node(
        ValidateImage(),
        name='val_bold',
        mem_gb=DEFAULT_MEMORY_MIN_GB,
    )

    # Drop frames to avoid startle when MRI begins acquiring
    select_frames = pe.Node(
        niu.Function(function=_select_frames, output_names=['start_frame', 't_mask']),
        name='select_frames',
    )
    select_frames.inputs.ref_frame_start = ref_frame_start

    get_dummy = pe.Node(NonsteadyStatesDetector(), name='get_dummy')
    gen_avg = pe.Node(RobustAverage(), name='gen_avg', mem_gb=1)

    calc_dummy_scans = pe.Node(
        niu.Function(function=pass_dummy_scans, output_names=['skip_vols_num']),
        name='calc_dummy_scans',
        run_without_submitting=True,
        mem_gb=DEFAULT_MEMORY_MIN_GB,
    )

    workflow.connect([
        (inputnode, val_bold, [('bold_file', 'in_file')]),
        (inputnode, get_dummy, [('bold_file', 'in_file')]),
        (inputnode, calc_dummy_scans, [('dummy_scans', 'dummy_scans')]),
        (val_bold, gen_avg, [('out_file', 'in_file')]),
        (val_bold, select_frames, [('out_file', 'in_file')]),
        (inputnode, select_frames, [('dummy_scans', 'dummy_scans')]),
        (select_frames, gen_avg, [('t_mask', 't_mask')]),
        (get_dummy, calc_dummy_scans, [('n_dummy', 'algo_dummy_scans')]),
        (val_bold, outputnode, [
            ('out_file', 'bold_file'),
            ('out_report', 'validation_report'),
        ]),
        (calc_dummy_scans, outputnode, [('skip_vols_num', 'skip_vols')]),
        (gen_avg, outputnode, [('out_file', 'boldref')]),
        (get_dummy, outputnode, [('n_dummy', 'algo_dummy_scans')]),
    ])  # fmt:skip

    return workflow


def _select_frames(
    in_file: str, ref_frame_start: int, dummy_scans: int | None
) -> tuple[int, list]:
    import warnings

    import nibabel as nb
    import numpy as np

    img = nb.load(in_file)
    img_len = img.shape[3]

    # Ensure start index is the largest of the two
    # Will usually be `ref_frame_start`
    start_frame = max(ref_frame_start, dummy_scans) if dummy_scans else ref_frame_start

    if start_frame >= img_len:
        warnings.warn(
            f'Caculating the BOLD reference starting on frame {start_frame} but only {img_len} '
            'volumes in BOLD file, so using last volume.',
            stacklevel=1,
        )
        start_frame = img_len - 1

    t_mask = np.array([False] * img_len, dtype=bool)
    t_mask[start_frame:] = True
    return start_frame, list(t_mask)



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
#
# Copyright The NiPreps Developers <nipreps@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# We support and encourage derived works from this project, please read
# about our expectations at
#
#     https://www.nipreps.org/community/licensing/
#
"""Utilities and mocks for testing and documentation building."""

import os
import shutil
from contextlib import contextmanager
from pathlib import Path
from tempfile import mkdtemp

from toml import loads

from nibabies import data
from nibabies.workflows.base import init_execution_spaces


@contextmanager
def mock_config(bids_dir=None):
    """Create a mock config for documentation and testing purposes."""
    from ... import config

    _old_fs = os.getenv('FREESURFER_HOME')
    if not _old_fs:
        os.environ['FREESURFER_HOME'] = mkdtemp()

    settings = loads(data.load.readable('tests/config.toml').read_text())
    for sectionname, configs in settings.items():
        if sectionname != 'environment':
            section = getattr(config, sectionname)
            section.load(configs, init=False)
    config.nipype.omp_nthreads = 1
    config.nipype.init()
    config.loggers.init()
    init_execution_spaces()

    bids_dir = bids_dir or data.load('tests/ds000005').absolute()

    config.execution.work_dir = Path(mkdtemp())
    config.execution.bids_dir = bids_dir
    config.execution.nibabies_dir = Path(mkdtemp())
    config.execution.bids_database_dir = None
    config.execution._layout = None
    config.execution.init()

    yield

    shutil.rmtree(config.execution.work_dir)
    shutil.rmtree(config.execution.nibabies_dir)

    if not _old_fs:
        del os.environ['FREESURFER_HOME']



import os
from argparse import ArgumentParser

import nipype.pipeline.engine as pe
from niworkflows.interfaces.nibabel import MapLabels

from nibabies.interfaces.mcribs import MCRIBReconAll


def _parser():
    parser = ArgumentParser(description='Test script for MCRIBS surfaces')
    parser.add_argument('subject', help='Subject ID')
    parser.add_argument('t2w', type=os.path.abspath, help='Input T2w (radioisotropic)')
    parser.add_argument(
        'segmentation', type=os.path.abspath, help='Input anatomical segmentation in T2w space'
    )
    parser.add_argument(
        '--outdir', type=os.path.abspath, help='Output directory to persist MCRIBS output'
    )
    parser.add_argument('--nthreads', type=int, help='Number of threads to parallelize tasks')
    return parser


def main(argv: list = None):
    pargs = _parser().parse_args(argv)

    t2w_file = _check_file(pargs.t2w)
    seg_file = _check_file(pargs.segmentation)

    aseg2mcrib = {
        2: 51,
        3: 21,
        4: 49,
        5: 0,
        7: 17,
        8: 17,
        10: 43,
        11: 41,
        12: 47,
        13: 47,
        14: 0,
        15: 0,
        16: 19,
        17: 1,
        18: 3,
        26: 41,
        28: 45,
        31: 49,
        41: 52,
        42: 20,
        43: 50,
        44: 0,
        46: 18,
        47: 18,
        49: 42,
        50: 40,
        51: 46,
        52: 46,
        53: 2,
        54: 4,
        58: 40,
        60: 44,
        63: 50,
        253: 48,
    }
    map_labels = pe.Node(MapLabels(in_file=seg_file, mappings=aseg2mcrib), name='map_labels')

    recon = pe.Node(
        MCRIBReconAll(subject_id=pargs.subject, t2w_file=t2w_file), name='mcribs_recon'
    )
    if pargs.outdir:
        recon.inputs.outdir = pargs.outdir
    if pargs.nthreads:
        recon.inputs.nthreads = pargs.nthreads

    wf = pe.Workflow(f'MRA_{pargs.subject}')
    wf.connect(map_labels, 'out_file', recon, 'segmentation_file')
    wf.run()


def _check_file(fl: str) -> str:
    import nibabel as nb
    import numpy as np

    img = nb.load(fl)
    if len(img.shape) != 3:
        raise ValueError('Image {fl} is not 3 dimensional.')

    voxdims = img.header['pixdim'][1:4]
    if not np.allclose(voxdims, voxdims[1]):
        raise ValueError(f'Image {fl} is not isotropic: {voxdims}.')

    ornt = nb.io_orientation(img.affine)
    axcodes = nb.orientations.ornt2axcodes(ornt)
    if ''.join(axcodes) != 'LAS':
        las = nb.orientations.axcodes2ornt('LAS')
        transform = nb.orientations.ornt_transform(ornt, las)
        reornt = img.as_reoriented(transform)
        outfl = os.path.abspath(f'LASornt_{os.path.basename(fl)}')
        print(f'Creating reorientated image {outfl}')
        reornt.to_filename(outfl)
        return outfl
    return fl


if __name__ == '__main__':
    main()



from pathlib import Path
from uuid import uuid4

import nibabel as nb
import numpy as np
import pytest

from ..nibabel import ReorientImage


def create_save_img(ornt: str):
    data = np.random.rand(2, 2, 2)
    img = nb.Nifti1Image(data, affine=np.eye(4))
    # img will always be in RAS at the start
    ras = nb.orientations.axcodes2ornt('RAS')
    if ornt != 'RAS':
        new = nb.orientations.axcodes2ornt(ornt)
        xfm = nb.orientations.ornt_transform(ras, new)
        img = img.as_reoriented(xfm)
    out_file = f'{uuid4()}.nii.gz'
    img.to_filename(out_file)
    return out_file


@pytest.mark.parametrize(
    ('in_ornt', 'out_ornt'),
    [
        ('RAS', 'RAS'),
        ('RAS', 'LAS'),
        ('LAS', 'RAS'),
        ('RAS', 'RPI'),
        ('LPI', 'RAS'),
    ],
)
def test_reorient_image(tmpdir, in_ornt, out_ornt):
    tmpdir.chdir()

    in_file = create_save_img(ornt=in_ornt)
    in_img = nb.load(in_file)
    assert ''.join(nb.aff2axcodes(in_img.affine)) == in_ornt

    # test string representation
    res = ReorientImage(in_file=in_file, target_orientation=out_ornt).run()
    out_file = res.outputs.out_file
    out_img = nb.load(out_file)
    assert ''.join(nb.aff2axcodes(out_img.affine)) == out_ornt
    Path(out_file).unlink()

    # test with target file
    target_file = create_save_img(ornt=out_ornt)
    target_img = nb.load(target_file)
    assert ''.join(nb.aff2axcodes(target_img.affine)) == out_ornt
    res = ReorientImage(in_file=in_file, target_file=target_file).run()
    out_file = res.outputs.out_file
    out_img = nb.load(out_file)
    assert ''.join(nb.aff2axcodes(out_img.affine)) == out_ornt

    # cleanup
    for f in (in_file, target_file, out_file):
        Path(f).unlink()
